from pathlib import Path

# we should make this a proper package that loads examples
# similar how they handle it in xarray:
# https://github.com/pydata/xarray/blob/main/xarray/tutorial.py
# Currently there are simple paths to the examples which are then easily loaded in the notebooks

_EXAMPLES = Path(__file__).parent
wind_energy_ontology = _EXAMPLES / "wind-energy.owl"



from collections.abc import Sequence
from dataclasses import dataclass
from datetime import datetime
from typing import Any, Literal, cast, overload

from rdflib import Namespace

from cognite.neat.rules.models._base import DataModelType, ExtensionCategory, SchemaCompleteness, _add_alias
from cognite.neat.rules.models.data_types import DataType
from cognite.neat.rules.models.entities import ClassEntity, ParentClassEntity, Unknown, UnknownEntity

from ._rules import InformationClass, InformationMetadata, InformationProperty, InformationRules


@dataclass
class InformationMetadataInput:
    schema_: Literal["complete", "partial", "extended"]
    prefix: str
    namespace: str
    version: str
    creator: str
    data_model_type: Literal["solution", "enterprise"] = "enterprise"
    extension: Literal["addition", "reshape", "rebuild"] = "addition"
    name: str | None = None
    description: str | None = None
    created: datetime | str | None = None
    updated: datetime | str | None = None
    license: str | None = None
    rights: str | None = None

    @classmethod
    def load(cls, data: dict[str, Any] | None) -> "InformationMetadataInput | None":
        if data is None:
            return None
        _add_alias(data, InformationMetadata)
        return cls(
            data_model_type=data.get("data_model_type", "enterprise"),
            extension=data.get("extension", "addition"),
            schema_=data.get("schema_", "partial"),  # type: ignore[arg-type]
            version=data.get("version"),  # type: ignore[arg-type]
            namespace=data.get("namespace"),  # type: ignore[arg-type]
            prefix=data.get("prefix"),  # type: ignore[arg-type]
            name=data.get("name"),
            creator=data.get("creator"),  # type: ignore[arg-type]
            description=data.get("description"),
            created=data.get("created"),
            updated=data.get("updated"),
            license=data.get("license"),
            rights=data.get("rights"),
        )

    def dump(self) -> dict[str, Any]:
        return dict(
            dataModelType=DataModelType(self.data_model_type),
            schema=SchemaCompleteness(self.schema_),
            extension=ExtensionCategory(self.extension),
            namespace=Namespace(self.namespace),
            prefix=self.prefix,
            version=self.version,
            name=self.name,
            creator=self.creator,
            description=self.description,
            created=self.created or datetime.now(),
            updated=self.updated or datetime.now(),
        )


@dataclass
class InformationPropertyInput:
    class_: str
    property_: str
    value_type: str
    name: str | None = None
    description: str | None = None
    comment: str | None = None
    min_count: int | None = None
    max_count: int | float | None = None
    default: Any | None = None
    reference: str | None = None
    match_type: str | None = None
    rule_type: str | None = None
    rule: str | None = None

    @classmethod
    @overload
    def load(cls, data: None) -> None: ...

    @classmethod
    @overload
    def load(cls, data: dict[str, Any]) -> "InformationPropertyInput": ...

    @classmethod
    @overload
    def load(cls, data: list[dict[str, Any]]) -> list["InformationPropertyInput"]: ...

    @classmethod
    def load(
        cls, data: dict[str, Any] | list[dict[str, Any]] | None
    ) -> "InformationPropertyInput | list[InformationPropertyInput] | None":
        if data is None:
            return None
        if isinstance(data, list) or (isinstance(data, dict) and isinstance(data.get("data"), list)):
            items = cast(list[dict[str, Any]], data.get("data") if isinstance(data, dict) else data)
            return [loaded for item in items if (loaded := cls.load(item)) is not None]

        _add_alias(data, InformationProperty)
        return cls(
            class_=data.get("class_"),  # type: ignore[arg-type]
            property_=data.get("property_"),  # type: ignore[arg-type]
            name=data.get("name", None),
            description=data.get("description", None),
            comment=data.get("comment", None),
            value_type=data.get("value_type"),  # type: ignore[arg-type]
            min_count=data.get("min_count", None),
            max_count=data.get("max_count", None),
            default=data.get("default", None),
            reference=data.get("reference", None),
            match_type=data.get("match_type", None),
            rule_type=data.get("rule_type", None),
            rule=data.get("rule", None),
        )

    def dump(self, default_prefix: str) -> dict[str, Any]:
        value_type: DataType | ClassEntity | UnknownEntity

        # property holding xsd data type
        if DataType.is_data_type(self.value_type):
            value_type = DataType.load(self.value_type)

        # unknown value type
        elif self.value_type == str(Unknown):
            value_type = UnknownEntity()

        # property holding link to class
        else:
            value_type = ClassEntity.load(self.value_type, prefix=default_prefix)

        return {
            "Class": ClassEntity.load(self.class_, prefix=default_prefix),
            "Property": self.property_,
            "Name": self.name,
            "Description": self.description,
            "Comment": self.comment,
            "Value Type": value_type,
            "Min Count": self.min_count,
            "Max Count": self.max_count,
            "Default": self.default,
            "Reference": self.reference,
            "Match Type": self.match_type,
            "Rule Type": self.rule_type,
            "Rule": self.rule,
        }


@dataclass
class InformationClassInput:
    class_: str
    name: str | None = None
    description: str | None = None
    comment: str | None = None
    parent: str | None = None
    reference: str | None = None
    match_type: str | None = None

    @classmethod
    @overload
    def load(cls, data: None) -> None: ...

    @classmethod
    @overload
    def load(cls, data: dict[str, Any]) -> "InformationClassInput": ...

    @classmethod
    @overload
    def load(cls, data: list[dict[str, Any]]) -> list["InformationClassInput"]: ...

    @classmethod
    def load(
        cls, data: dict[str, Any] | list[dict[str, Any]] | None
    ) -> "InformationClassInput | list[InformationClassInput] | None":
        if data is None:
            return None
        if isinstance(data, list) or (isinstance(data, dict) and isinstance(data.get("data"), list)):
            items = cast(list[dict[str, Any]], data.get("data") if isinstance(data, dict) else data)
            return [loaded for item in items if (loaded := cls.load(item)) is not None]
        _add_alias(data, InformationClass)
        return cls(
            class_=data.get("class_"),  # type: ignore[arg-type]
            name=data.get("name", None),
            description=data.get("description", None),
            comment=data.get("comment", None),
            parent=data.get("parent", None),
            reference=data.get("reference", None),
            match_type=data.get("match_type", None),
        )

    def dump(self, default_prefix: str) -> dict[str, Any]:
        return {
            "Class": ClassEntity.load(self.class_, prefix=default_prefix),
            "Name": self.name,
            "Description": self.description,
            "Comment": self.comment,
            "Reference": self.reference,
            "Match Type": self.match_type,
            "Parent Class": (
                [ParentClassEntity.load(parent, prefix=default_prefix) for parent in self.parent.split(",")]
                if self.parent
                else None
            ),
        }


@dataclass
class InformationRulesInput:
    metadata: InformationMetadataInput
    properties: Sequence[InformationPropertyInput]
    classes: Sequence[InformationClassInput]
    last: "InformationRulesInput | InformationRules | None" = None
    reference: "InformationRulesInput | InformationRules | None" = None

    @classmethod
    @overload
    def load(cls, data: dict[str, Any]) -> "InformationRulesInput": ...

    @classmethod
    @overload
    def load(cls, data: None) -> None: ...

    @classmethod
    def load(cls, data: dict | None) -> "InformationRulesInput | None":
        if data is None:
            return None
        _add_alias(data, InformationRules)

        return cls(
            metadata=InformationMetadataInput.load(data.get("metadata")),  # type: ignore[arg-type]
            properties=InformationPropertyInput.load(data.get("properties")),  # type: ignore[arg-type]
            classes=InformationClassInput.load(data.get("classes")),  # type: ignore[arg-type]
            last=InformationRulesInput.load(data.get("last")),
            reference=InformationRulesInput.load(data.get("reference")),
        )

    def as_rules(self) -> InformationRules:
        return InformationRules.model_validate(self.dump())

    def dump(self) -> dict[str, Any]:
        default_prefix = self.metadata.prefix
        reference: dict[str, Any] | None = None
        if isinstance(self.reference, InformationRulesInput):
            reference = self.reference.dump()
        elif isinstance(self.reference, InformationRules):
            # We need to load through the InformationRulesInput to set the correct default space and version
            reference = InformationRulesInput.load(self.reference.model_dump()).dump()
        last: dict[str, Any] | None = None
        if isinstance(self.last, InformationRulesInput):
            last = self.last.dump()
        elif isinstance(self.last, InformationRules):
            # We need to load through the InformationRulesInput to set the correct default space and version
            last = InformationRulesInput.load(self.last.model_dump()).dump()
        return dict(
            Metadata=self.metadata.dump(),
            Properties=[prop.dump(default_prefix) for prop in self.properties],
            Classes=[class_.dump(default_prefix) for class_ in self.classes],
            Last=last,
            Reference=reference,
        )



import json
import sys
import warnings
import zipfile
from collections import Counter, defaultdict
from dataclasses import Field, dataclass, field, fields
from pathlib import Path
from typing import Any, ClassVar, cast

import yaml
from cognite.client import CogniteClient
from cognite.client import data_modeling as dm
from cognite.client.data_classes import DatabaseWrite, DatabaseWriteList, TransformationWrite, TransformationWriteList
from cognite.client.data_classes.data_modeling import ViewApply
from cognite.client.data_classes.data_modeling.views import ReverseDirectRelation
from cognite.client.data_classes.transformations.common import Edges, EdgeType, Nodes, ViewInfo

from cognite.neat.rules import issues
from cognite.neat.rules.issues.dms import (
    ContainerPropertyUsedMultipleTimesError,
    DirectRelationMissingSourceWarning,
    DMSSchemaError,
    DuplicatedViewInDataModelError,
    IncompleteSchemaError,
    MissingContainerError,
    MissingContainerPropertyError,
    MissingEdgeViewError,
    MissingParentViewError,
    MissingSourceViewError,
    MissingSpaceError,
    MissingViewError,
    MissingViewInModelWarning,
)
from cognite.neat.rules.models.data_types import _DATA_TYPE_BY_DMS_TYPE
from cognite.neat.utils.cdf_classes import (
    CogniteResourceDict,
    ContainerApplyDict,
    NodeApplyDict,
    SpaceApplyDict,
    ViewApplyDict,
)
from cognite.neat.utils.cdf_loaders import ViewLoader
from cognite.neat.utils.cdf_loaders.data_classes import RawTableWrite, RawTableWriteList
from cognite.neat.utils.text import to_camel
from cognite.neat.utils.utils import get_inheritance_path

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


@dataclass
class DMSSchema:
    data_model: dm.DataModelApply | None = None
    spaces: SpaceApplyDict = field(default_factory=SpaceApplyDict)
    views: ViewApplyDict = field(default_factory=ViewApplyDict)
    containers: ContainerApplyDict = field(default_factory=ContainerApplyDict)
    node_types: NodeApplyDict = field(default_factory=NodeApplyDict)
    # The last schema is the previous version of the data model. In the case, extension=addition, this
    # should not be modified.
    last: "DMSSchema | None" = None
    # Reference is typically the Enterprise model, while this is the solution model.
    reference: "DMSSchema | None" = None

    _FIELD_NAME_BY_RESOURCE_TYPE: ClassVar[dict[str, str]] = {
        "container": "containers",
        "view": "views",
        "datamodel": "data_model",
        "space": "spaces",
        "node": "node_types",
    }

    def _get_mapped_container_from_view(self, view_id: dm.ViewId) -> set[dm.ContainerId]:
        # index all views, including ones from reference
        view_by_id = self.views.copy()
        if self.reference:
            view_by_id.update(self.reference.views)

        if view_id not in view_by_id:
            raise ValueError(f"View {view_id} not found")

        indexed_implemented_views = {id_: view.implements for id_, view in view_by_id.items()}
        view_inheritance = get_inheritance_path(view_id, indexed_implemented_views)

        directly_referenced_containers = view_by_id[view_id].referenced_containers()
        inherited_referenced_containers = set()

        for view_id in view_inheritance:
            if implemented_view := view_by_id.get(view_id):
                inherited_referenced_containers |= implemented_view.referenced_containers()
            else:
                raise IncompleteSchemaError(missing_component=view_id).as_exception()

        return directly_referenced_containers | inherited_referenced_containers

    @classmethod
    def from_model_id(cls, client: CogniteClient, data_model_id: dm.DataModelIdentifier) -> "DMSSchema":
        data_models = client.data_modeling.data_models.retrieve(data_model_id, inline_views=True)
        if len(data_models) == 0:
            raise ValueError(f"Data model {data_model_id} not found")
        data_model = data_models.latest_version()
        return cls.from_data_model(client, data_model)

    @classmethod
    def from_data_model(
        cls,
        client: CogniteClient,
        data_model: dm.DataModel[dm.View],
        reference_model: dm.DataModel[dm.View] | None = None,
    ) -> "DMSSchema":
        """Create a schema from a data model.

        If a reference model is provided, the schema will include a reference schema. To determine which views,
        and containers to put in the reference schema, the following rule is applied:

            If a view or container space is different from the data model space,
            it will be included in the reference schema.*

        *One exception to this rule is if a view is directly referenced by the data model. In this case, the view will
        be included in the data model schema, even if the space is different.

        Args:
            client: The Cognite client used for retrieving components referenced by the data model.
            data_model: The data model to create the schema from.
            reference_model: (Optional) The reference model to include in the schema.
                This is typically the Enterprise model.

        Returns:
            DMSSchema: The schema created from the data model.
        """
        views = dm.ViewList(data_model.views)

        data_model_write = data_model.as_write()
        data_model_write.views = list(views.as_ids())

        if reference_model:
            views.extend(reference_model.views)

        container_ids = views.referenced_containers()
        containers = client.data_modeling.containers.retrieve(list(container_ids))
        cls._append_referenced_containers(client, containers)

        space_ids = [data_model.space, reference_model.space] if reference_model else [data_model.space]
        space_read = client.data_modeling.spaces.retrieve(space_ids)
        if len(space_read) != len(space_ids):
            raise ValueError(f"Space(s) {space_read} not found")
        space_write = space_read.as_write()

        view_loader = ViewLoader(client)

        existing_view_ids = set(views.as_ids())

        # We need to include all views the edges/direct relations are pointing to have a complete schema.
        connection_referenced_view_ids: set[dm.ViewId] = set()
        for view in views:
            connection_referenced_view_ids |= cls._connection_references(view)
        connection_referenced_view_ids = connection_referenced_view_ids - existing_view_ids
        if connection_referenced_view_ids:
            warnings.warn(
                MissingViewInModelWarning(data_model.as_id(), connection_referenced_view_ids), UserWarning, stacklevel=2
            )
            connection_referenced_views = view_loader.retrieve(list(connection_referenced_view_ids))
            if failed := connection_referenced_view_ids - set(connection_referenced_views.as_ids()):
                warnings.warn(
                    issues.importing.FailedImportWarning({repr(v) for v in failed}), UserWarning, stacklevel=2
                )
            views.extend(connection_referenced_views)

        # We need to include parent views in the schema to make sure that the schema is valid.
        parent_view_ids = {parent for view in views for parent in view.implements or []}
        parents = view_loader.retrieve_all_parents(list(parent_view_ids - existing_view_ids))
        views.extend([parent for parent in parents if parent.as_id() not in existing_view_ids])

        # Converting views from read to write format requires to account for parents (implements)
        # as the read format contains all properties from all parents, while the write formate should not contain
        # properties from any parents.
        # The ViewLoader as_write method looks up parents and remove properties from them.
        view_write = ViewApplyDict([view_loader.as_write(view) for view in views])

        container_write = ContainerApplyDict(containers.as_write())
        user_space = data_model.space
        if reference_model:
            user_model_view_ids = set(data_model_write.views)
            ref_model_write = reference_model.as_write()
            ref_model_write.views = [view.as_id() for view in reference_model.views]

            ref_views = ViewApplyDict(
                [
                    view
                    for view_id, view in view_write.items()
                    if (view.space != user_space) or (view_id not in user_model_view_ids)
                ]
            )
            view_write = ViewApplyDict(
                [
                    view
                    for view_id, view in view_write.items()
                    if view.space == user_space or view_id in user_model_view_ids
                ]
            )

            ref_containers = ContainerApplyDict(
                [container for container in container_write.values() if container.space != user_space]
            )
            container_write = ContainerApplyDict(
                [container for container in container_write.values() if container.space == user_space]
            )

            ref_schema: DMSSchema | None = cls(
                spaces=SpaceApplyDict([s for s in space_write if s.space != user_space]),
                data_model=ref_model_write,
                views=ref_views,
                containers=ref_containers,
            )
        else:
            ref_schema = None
        return cls(
            spaces=SpaceApplyDict([s for s in space_write if s.space == user_space]),
            data_model=data_model_write,
            views=view_write,
            containers=container_write,
            reference=ref_schema,
        )

    @classmethod
    def _connection_references(cls, view: dm.View) -> set[dm.ViewId]:
        view_ids: set[dm.ViewId] = set()
        for prop in (view.properties or {}).values():
            if isinstance(prop, dm.MappedProperty) and isinstance(prop.type, dm.DirectRelation):
                if prop.source:
                    view_ids.add(prop.source)
            elif isinstance(prop, dm.EdgeConnection):
                view_ids.add(prop.source)
                if prop.edge_source:
                    view_ids.add(prop.edge_source)
            elif isinstance(prop, ReverseDirectRelation):
                view_ids.add(prop.source)
        return view_ids

    @classmethod
    def from_directory(cls, directory: str | Path) -> Self:
        """Load a schema from a directory containing YAML files.

        The directory is expected to follow the Cognite-Toolkit convention
        where each file is named as `resource_type.resource_name.yaml`.
        """
        data, context = cls._read_directory(Path(directory))
        return cls.load(data, context)

    @classmethod
    def _read_directory(cls, directory: Path) -> tuple[dict[str, list[Any]], dict[str, list[Path]]]:
        data: dict[str, Any] = {}
        context: dict[str, list[Path]] = {}
        for yaml_file in directory.rglob("*.yaml"):
            if "." in yaml_file.stem:
                resource_type = yaml_file.stem.rsplit(".", 1)[-1]
                if attr_name := cls._FIELD_NAME_BY_RESOURCE_TYPE.get(resource_type):
                    data.setdefault(attr_name, [])
                    context.setdefault(attr_name, [])
                    try:
                        # Using CSafeLoader over safe_load for ~10x speedup
                        loaded = yaml.safe_load(yaml_file.read_text())
                    except Exception as e:
                        warnings.warn(issues.fileread.InvalidFileFormatWarning(yaml_file, str(e)), stacklevel=2)
                        continue

                    if isinstance(loaded, list):
                        data[attr_name].extend(loaded)
                        context[attr_name].extend([yaml_file] * len(loaded))
                    else:
                        data[attr_name].append(loaded)
                        context[attr_name].append(yaml_file)
        return data, context

    def to_directory(
        self,
        directory: str | Path,
        exclude: set[str] | None = None,
        new_line: str | None = "\n",
        encoding: str | None = "utf-8",
    ) -> None:
        """Save the schema to a directory as YAML files. This is compatible with the Cognite-Toolkit convention.

        Args:
            directory (str | Path): The directory to save the schema to.
            exclude (set[str]): A set of attributes to exclude from the output.
            new_line (str): The line endings to use in the output files. Defaults to "\n".
            encoding (str): The encoding to use in the output files. Defaults to "utf-8".
        """
        path_dir = Path(directory)
        exclude_set = exclude or set()
        data_models = path_dir / "data_models"
        data_models.mkdir(parents=True, exist_ok=True)
        if "spaces" not in exclude_set:
            for space in self.spaces.values():
                (data_models / f"{space.space}.space.yaml").write_text(
                    space.dump_yaml(), newline=new_line, encoding=encoding
                )
        if "data_models" not in exclude_set and self.data_model:
            (data_models / f"{self.data_model.external_id}.datamodel.yaml").write_text(
                self.data_model.dump_yaml(), newline=new_line, encoding=encoding
            )
        if "views" not in exclude_set and self.views:
            view_dir = data_models / "views"
            view_dir.mkdir(parents=True, exist_ok=True)
            for view in self.views.values():
                (view_dir / f"{view.external_id}.view.yaml").write_text(
                    view.dump_yaml(), newline=new_line, encoding=encoding
                )
        if "containers" not in exclude_set and self.containers:
            container_dir = data_models / "containers"
            container_dir.mkdir(parents=True, exist_ok=True)
            for container in self.containers.values():
                (container_dir / f"{container.external_id}.container.yaml").write_text(
                    container.dump_yaml(), newline=new_line, encoding=encoding
                )
        if "node_types" not in exclude_set and self.node_types:
            node_dir = data_models / "nodes"
            node_dir.mkdir(parents=True, exist_ok=True)
            for node in self.node_types.values():
                (node_dir / f"{node.external_id}.node.yaml").write_text(
                    node.dump_yaml(), newline=new_line, encoding=encoding
                )

    @classmethod
    def from_zip(cls, zip_file: str | Path) -> Self:
        """Load a schema from a ZIP file containing YAML files.

        The ZIP file is expected to follow the Cognite-Toolkit convention
        where each file is named as `resource_type.resource_name.yaml`.
        """
        data, context = cls._read_zip(Path(zip_file))
        return cls.load(data, context)

    @classmethod
    def _read_zip(cls, zip_file: Path) -> tuple[dict[str, list[Any]], dict[str, list[Path]]]:
        data: dict[str, list[Any]] = {}
        context: dict[str, list[Path]] = {}
        with zipfile.ZipFile(zip_file, "r") as zip_ref:
            for file_info in zip_ref.infolist():
                if file_info.filename.endswith(".yaml"):
                    if "/" not in file_info.filename:
                        continue
                    filename = Path(file_info.filename.split("/")[-1])
                    if "." not in filename.stem:
                        continue
                    resource_type = filename.stem.rsplit(".", 1)[-1]
                    if attr_name := cls._FIELD_NAME_BY_RESOURCE_TYPE.get(resource_type):
                        data.setdefault(attr_name, [])
                        context.setdefault(attr_name, [])
                        try:
                            loaded = yaml.safe_load(zip_ref.read(file_info).decode())
                        except Exception as e:
                            warnings.warn(issues.fileread.InvalidFileFormatWarning(filename, str(e)), stacklevel=2)
                            continue
                        if isinstance(loaded, list):
                            data[attr_name].extend(loaded)
                            context[attr_name].extend([filename] * len(loaded))
                        else:
                            data[attr_name].append(loaded)
                            context[attr_name].append(filename)
        return data, context

    def to_zip(self, zip_file: str | Path, exclude: set[str] | None = None) -> None:
        """Save the schema to a ZIP file as YAML files. This is compatible with the Cognite-Toolkit convention.

        Args:
            zip_file (str | Path): The ZIP file to save the schema to.
            exclude (set[str]): A set of attributes to exclude from the output.
        """
        exclude_set = exclude or set()
        with zipfile.ZipFile(zip_file, "w") as zip_ref:
            if "spaces" not in exclude_set:
                for space in self.spaces.values():
                    zip_ref.writestr(f"data_models/{space.space}.space.yaml", space.dump_yaml())
            if "data_models" not in exclude_set and self.data_model:
                zip_ref.writestr(
                    f"data_models/{self.data_model.external_id}.datamodel.yaml", self.data_model.dump_yaml()
                )
            if "views" not in exclude_set:
                for view in self.views.values():
                    zip_ref.writestr(f"data_models/views/{view.external_id}.view.yaml", view.dump_yaml())
            if "containers" not in exclude_set:
                for container in self.containers.values():
                    zip_ref.writestr(
                        f"data_models/containers{container.external_id}.container.yaml", container.dump_yaml()
                    )
            if "node_types" not in exclude_set:
                for node in self.node_types.values():
                    zip_ref.writestr(f"data_models/nodes/{node.external_id}.node.yaml", node.dump_yaml())

    @classmethod
    def load(cls, data: str | dict[str, list[Any]], context: dict[str, list[Path]] | None = None) -> Self:
        """Loads a schema from a dictionary or a YAML or JSON formatted string.

        Args:
            data: The data to load the schema from. This can be a dictionary, a YAML or JSON formatted string.
            context: This provides linage for where the data was loaded from. This is used in Warnings
                if a single item fails to load.

        Returns:
            DMSSchema: The loaded schema.
        """
        context = context or {}
        if isinstance(data, str):
            # YAML is a superset of JSON, so we can use the same parser
            try:
                data_dict = yaml.safe_load(data)
            except Exception as e:
                raise issues.fileread.FailedStringLoadError(".yaml", str(e)).as_exception() from None
            if not isinstance(data_dict, dict) and all(isinstance(v, list) for v in data_dict.values()):
                raise issues.fileread.FailedStringLoadError(
                    "dict[str, list[Any]]", f"Invalid data structure: {type(data)}"
                ).as_exception() from None
        else:
            data_dict = data
        loaded: dict[str, Any] = {}
        for attr in fields(cls):
            if items := data_dict.get(attr.name) or data_dict.get(to_camel(attr.name)):
                if attr.name == "data_model":
                    if isinstance(items, list) and len(items) > 1:
                        warnings.warn(
                            issues.importing.MultipleDataModelsWarning(
                                [item.get("externalId", "Unknown") for item in items]
                            ),
                            stacklevel=2,
                        )
                    item = items[0] if isinstance(items, list) else items
                    try:
                        loaded[attr.name] = dm.DataModelApply.load(item)
                    except Exception as e:
                        data_model_file = context.get(attr.name, [Path("UNKNOWN")])[0]
                        warnings.warn(
                            issues.fileread.FailedLoadWarning(data_model_file, dm.DataModelApply.__name__, str(e)),
                            stacklevel=2,
                        )
                else:
                    try:
                        loaded[attr.name] = attr.type.load(items)
                    except Exception as e:
                        loaded[attr.name] = cls._load_individual_resources(
                            items, attr, str(e), context.get(attr.name, [])
                        )
        return cls(**loaded)

    @classmethod
    def _load_individual_resources(cls, items: list, attr: Field, trigger_error: str, resource_context) -> list[Any]:
        resources = attr.type([])
        if not hasattr(attr.type, "_RESOURCE"):
            warnings.warn(
                issues.fileread.FailedLoadWarning(Path("UNKNOWN"), attr.type.__name__, trigger_error), stacklevel=2
            )
            return resources
        # Fallback to load individual resources.
        single_cls = attr.type._RESOURCE
        for no, item in enumerate(items):
            try:
                loaded_instance = single_cls.load(item)
            except Exception as e:
                try:
                    filepath = resource_context[no]
                except IndexError:
                    filepath = Path("UNKNOWN")
                # We use repr(e) instead of str(e) to include the exception type in the warning message
                warnings.warn(issues.fileread.FailedLoadWarning(filepath, single_cls.__name__, repr(e)), stacklevel=2)
            else:
                resources.append(loaded_instance)
        return resources

    def dump(self, camel_case: bool = True, sort: bool = True) -> dict[str, Any]:
        """Dump the schema to a dictionary that can be serialized to JSON.

        Args:
            camel_case (bool): If True, the keys in the output dictionary will be in camel case.
            sort (bool): If True, the items in the output dictionary will be sorted by their ID.
                This is useful for deterministic output which is useful for comparing schemas.

        Returns:
            dict: The schema as a dictionary.
        """
        output: dict[str, Any] = {}
        cls_fields = sorted(fields(self), key=lambda f: f.name) if sort else fields(self)
        for attr in cls_fields:
            if items := getattr(self, attr.name):
                key = to_camel(attr.name) if camel_case else attr.name
                if isinstance(items, CogniteResourceDict):
                    if sort:
                        output[key] = [
                            item.dump(camel_case) for item in sorted(items.values(), key=self._to_sortable_identifier)
                        ]
                    else:
                        output[key] = items.dump(camel_case)
                else:
                    output[key] = items.dump(camel_case=camel_case)
        return output

    @classmethod
    def _to_sortable_identifier(cls, item: Any) -> str | tuple[str, str] | tuple[str, str, str]:
        if isinstance(item, dm.ContainerApply | dm.ViewApply | dm.DataModelApply | dm.NodeApply | RawTableWrite):
            identifier = item.as_id().as_tuple()
            if len(identifier) == 3 and identifier[2] is None:
                return identifier[:2]  # type: ignore[misc]
            return cast(tuple[str, str] | tuple[str, str, str], identifier)
        elif isinstance(item, dm.SpaceApply):
            return item.space
        elif isinstance(item, TransformationWrite):
            return item.external_id or ""
        elif isinstance(item, DatabaseWrite):
            return item.name or ""
        else:
            raise ValueError(f"Cannot sort item of type {type(item)}")

    def validate(self) -> list[DMSSchemaError]:
        errors: set[DMSSchemaError] = set()
        defined_spaces = self.spaces.copy()
        defined_containers = self.containers.copy()
        defined_views = self.views.copy()
        for other_schema in [self.reference, self.last]:
            if other_schema:
                defined_spaces |= other_schema.spaces
                defined_containers |= other_schema.containers
                defined_views |= other_schema.views

        for container in self.containers.values():
            if container.space not in defined_spaces:
                errors.add(MissingSpaceError(space=container.space, referred_by=container.as_id()))

        for view in self.views.values():
            view_id = view.as_id()
            if view.space not in defined_spaces:
                errors.add(MissingSpaceError(space=view.space, referred_by=view_id))

            for parent in view.implements or []:
                if parent not in defined_views:
                    errors.add(MissingParentViewError(view=parent, referred_by=view_id))

            for prop_name, prop in (view.properties or {}).items():
                if isinstance(prop, dm.MappedPropertyApply):
                    ref_container = defined_containers.get(prop.container)
                    if ref_container is None:
                        errors.add(MissingContainerError(container=prop.container, referred_by=view_id))
                    elif prop.container_property_identifier not in ref_container.properties:
                        errors.add(
                            MissingContainerPropertyError(
                                container=prop.container,
                                property=prop.container_property_identifier,
                                referred_by=view_id,
                            )
                        )
                    else:
                        container_property = ref_container.properties[prop.container_property_identifier]

                        if isinstance(container_property.type, dm.DirectRelation) and prop.source is None:
                            warnings.warn(
                                DirectRelationMissingSourceWarning(view_id=view_id, property=prop_name), stacklevel=2
                            )

                if isinstance(prop, dm.EdgeConnectionApply) and prop.source not in defined_views:
                    errors.add(MissingSourceViewError(view=prop.source, property=prop_name, referred_by=view_id))

                if (
                    isinstance(prop, dm.EdgeConnectionApply)
                    and prop.edge_source is not None
                    and prop.edge_source not in defined_views
                ):
                    errors.add(MissingEdgeViewError(view=prop.edge_source, property=prop_name, referred_by=view_id))

            # This allows for multiple view properties to be mapped to the same container property,
            # as long as they have different external_id, otherwise this will lead to raising
            # error ContainerPropertyUsedMultipleTimesError
            property_count = Counter(
                (prop.container, prop.container_property_identifier, view_property_identifier)
                for view_property_identifier, prop in (view.properties or {}).items()
                if isinstance(prop, dm.MappedPropertyApply)
            )

            for (
                container_id,
                container_property_identifier,
                _,
            ), count in property_count.items():
                if count > 1:
                    view_properties = [
                        prop_name
                        for prop_name, prop in (view.properties or {}).items()
                        if isinstance(prop, dm.MappedPropertyApply)
                        and (prop.container, prop.container_property_identifier)
                        == (container_id, container_property_identifier)
                    ]
                    errors.add(
                        ContainerPropertyUsedMultipleTimesError(
                            container=container_id,
                            property=container_property_identifier,
                            referred_by=frozenset({(view_id, prop_name) for prop_name in view_properties}),
                        )
                    )

        if self.data_model:
            model = self.data_model
            if model.space not in defined_spaces:
                errors.add(MissingSpaceError(space=model.space, referred_by=model.as_id()))

            view_counts: dict[dm.ViewId, int] = defaultdict(int)
            for view_id_or_class in model.views or []:
                view_id = view_id_or_class if isinstance(view_id_or_class, dm.ViewId) else view_id_or_class.as_id()
                if view_id not in defined_views:
                    errors.add(MissingViewError(referred_by=model.as_id(), view=view_id))
                view_counts[view_id] += 1

            for view_id, count in view_counts.items():
                if count > 1:
                    errors.add(DuplicatedViewInDataModelError(referred_by=model.as_id(), view=view_id))

        return list(errors)

    @classmethod
    def _append_referenced_containers(cls, client: CogniteClient, containers: dm.ContainerList) -> None:
        """Containers can reference each other through the 'requires' constraint.

        This method retrieves all containers that are referenced by other containers through the 'requires' constraint,
        including their parents.

        """
        for _ in range(10):  # Limiting the number of iterations to avoid infinite loops
            referenced_containers = {
                const.require
                for container in containers
                for const in (container.constraints or {}).values()
                if isinstance(const, dm.RequiresConstraint)
            }
            missing_containers = referenced_containers - set(containers.as_ids())
            if not missing_containers:
                break
            found_containers = client.data_modeling.containers.retrieve(list(missing_containers))
            containers.extend(found_containers)
            if len(found_containers) != len(missing_containers):
                break
        else:
            warnings.warn(
                "The maximum number of iterations was reached while resolving referenced containers."
                "There might be referenced containers that are not included in the list of containers.",
                RuntimeWarning,
                stacklevel=2,
            )
        return None

    def referenced_spaces(self, include_indirect_references: bool = True) -> set[str]:
        """Get the spaces referenced by the schema.

        Args:
            include_indirect_references (bool): If True, the spaces referenced by as view.implements, and
                view.referenced_containers will be included in the output.
        Returns:
            set[str]: The spaces referenced by the schema.
        """
        referenced_spaces = {view.space for view in self.views.values()}
        referenced_spaces |= {container.space for container in self.containers.values()}
        if include_indirect_references:
            referenced_spaces |= {
                container.space for view in self.views.values() for container in view.referenced_containers()
            }
            referenced_spaces |= {parent.space for view in self.views.values() for parent in view.implements or []}
        referenced_spaces |= {node.space for node in self.node_types.values()}
        if self.data_model:
            referenced_spaces |= {self.data_model.space}
            referenced_spaces |= {view.space for view in self.data_model.views or []}
        referenced_spaces |= {s.space for s in self.spaces.values()}
        return referenced_spaces


@dataclass
class PipelineSchema(DMSSchema):
    transformations: TransformationWriteList = field(default_factory=lambda: TransformationWriteList([]))
    databases: DatabaseWriteList = field(default_factory=lambda: DatabaseWriteList([]))
    raw_tables: RawTableWriteList = field(default_factory=lambda: RawTableWriteList([]))

    _FIELD_NAME_BY_RESOURCE_TYPE: ClassVar[dict[str, str]] = {
        **DMSSchema._FIELD_NAME_BY_RESOURCE_TYPE,
        "raw": "raw_tables",
    }

    def __post_init__(self):
        existing_databases = {database.name for database in self.databases}
        table_database = {table.database for table in self.raw_tables}
        if missing := table_database - existing_databases:
            self.databases.extend([DatabaseWrite(name=database) for database in missing])

    @classmethod
    def _read_directory(cls, directory: Path) -> tuple[dict[str, list[Any]], dict[str, list[Path]]]:
        data, context = super()._read_directory(directory)
        for yaml_file in directory.rglob("*.yaml"):
            if yaml_file.parent.name in ("transformations", "raw"):
                attr_name = cls._FIELD_NAME_BY_RESOURCE_TYPE.get(yaml_file.parent.name, yaml_file.parent.name)
                data.setdefault(attr_name, [])
                context.setdefault(attr_name, [])
                try:
                    loaded = yaml.safe_load(yaml_file.read_text())
                except Exception as e:
                    warnings.warn(issues.fileread.InvalidFileFormatWarning(yaml_file, str(e)), stacklevel=2)
                    continue
                if isinstance(loaded, list):
                    data[attr_name].extend(loaded)
                    context[attr_name].extend([yaml_file] * len(loaded))
                else:
                    data[attr_name].append(loaded)
                    context[attr_name].append(yaml_file)
        return data, context

    def to_directory(
        self,
        directory: str | Path,
        exclude: set[str] | None = None,
        new_line: str | None = "\n",
        encoding: str | None = "utf-8",
    ) -> None:
        super().to_directory(directory, exclude)
        exclude_set = exclude or set()
        path_dir = Path(directory)
        if "transformations" not in exclude_set and self.transformations:
            transformation_dir = path_dir / "transformations"
            transformation_dir.mkdir(exist_ok=True, parents=True)
            for transformation in self.transformations:
                (transformation_dir / f"{transformation.external_id}.yaml").write_text(
                    transformation.dump_yaml(), newline=new_line, encoding=encoding
                )
        if "raw" not in exclude_set and self.raw_tables:
            # The RAW Databases are not written to file. This is because cognite-toolkit expects the RAW databases
            # to be in the same file as the RAW tables.
            raw_dir = path_dir / "raw"
            raw_dir.mkdir(exist_ok=True, parents=True)
            for raw_table in self.raw_tables:
                (raw_dir / f"{raw_table.name}.yaml").write_text(
                    raw_table.dump_yaml(), newline=new_line, encoding=encoding
                )

    def to_zip(self, zip_file: str | Path, exclude: set[str] | None = None) -> None:
        super().to_zip(zip_file, exclude)
        exclude_set = exclude or set()
        with zipfile.ZipFile(zip_file, "a") as zip_ref:
            if "transformations" not in exclude_set:
                for transformation in self.transformations:
                    zip_ref.writestr(f"transformations/{transformation.external_id}.yaml", transformation.dump_yaml())
            if "raw" not in exclude_set:
                # The RAW Databases are not written to file. This is because cognite-toolkit expects the RAW databases
                # to be in the same file as the RAW tables.
                for raw_table in self.raw_tables:
                    zip_ref.writestr(f"raw/{raw_table.name}.yaml", raw_table.dump_yaml())

    @classmethod
    def _read_zip(cls, zip_file: Path) -> tuple[dict[str, list[Any]], dict[str, list[Path]]]:
        data, context = super()._read_zip(zip_file)
        with zipfile.ZipFile(zip_file, "r") as zip_ref:
            for file_info in zip_ref.infolist():
                if file_info.filename.endswith(".yaml"):
                    if "/" not in file_info.filename:
                        continue
                    filepath = Path(file_info.filename)
                    if (parent := filepath.parent.name) in ("transformations", "raw"):
                        attr_name = cls._FIELD_NAME_BY_RESOURCE_TYPE.get(parent, parent)
                        data.setdefault(attr_name, [])
                        context.setdefault(attr_name, [])
                        try:
                            loaded = yaml.safe_load(zip_ref.read(file_info).decode())
                        except Exception as e:
                            warnings.warn(issues.fileread.InvalidFileFormatWarning(filepath, str(e)), stacklevel=2)
                            continue
                        if isinstance(loaded, list):
                            data[attr_name].extend(loaded)
                            context[attr_name].extend([filepath] * len(loaded))
                        else:
                            data[attr_name].append(loaded)
                            context[attr_name].append(filepath)
        return data, context

    @classmethod
    def from_dms(cls, schema: DMSSchema, instance_space: str | None = None) -> "PipelineSchema":
        if not schema.data_model:
            raise ValueError("PipelineSchema must contain at least one data model")
        first_data_model = schema.data_model
        # The database name is limited to 32 characters
        database_name = first_data_model.external_id[:32]
        instance_space = instance_space or first_data_model.space
        database = DatabaseWrite(name=database_name)
        parent_views = {parent for view in schema.views.values() for parent in view.implements or []}
        container_by_id = schema.containers.copy()

        transformations = TransformationWriteList([])
        raw_tables = RawTableWriteList([])
        for view in schema.views.values():
            if view.as_id() in parent_views:
                # Skipping parents as they do not have their own data
                continue
            mapped_properties = {
                prop_name: prop
                for prop_name, prop in (view.properties or {}).items()
                if isinstance(prop, dm.MappedPropertyApply)
            }
            if mapped_properties:
                view_table = RawTableWrite(name=f"{view.external_id}Properties", database=database_name)
                raw_tables.append(view_table)
                transformation = cls._create_property_transformation(
                    mapped_properties, view, view_table, container_by_id, instance_space
                )
                transformations.append(transformation)
            connection_properties = {
                prop_name: prop
                for prop_name, prop in (view.properties or {}).items()
                if isinstance(prop, dm.EdgeConnectionApply)
            }
            for prop_name, connection_property in connection_properties.items():
                view_table = RawTableWrite(name=f"{view.external_id}.{prop_name}Edge", database=database_name)
                raw_tables.append(view_table)
                transformation = cls._create_edge_transformation(connection_property, view, view_table, instance_space)
                transformations.append(transformation)

        return cls(
            spaces=schema.spaces,
            data_model=schema.data_model,
            views=schema.views,
            containers=schema.containers,
            transformations=transformations,
            databases=DatabaseWriteList([database]),
            raw_tables=raw_tables,
        )

    @classmethod
    def _create_property_transformation(
        cls,
        properties: dict[str, dm.MappedPropertyApply],
        view: ViewApply,
        table: RawTableWrite,
        container_by_id: dict[dm.ContainerId, dm.ContainerApply],
        instance_space: str,
    ) -> TransformationWrite:
        mapping_mode = {
            "version": 1,
            "sourceType": "raw",
            # 'mappings' is set here and overwritten further down to ensure the correct order
            "mappings": [],
            "sourceLevel1": table.database,
            "sourceLevel2": table.name,
        }
        mappings = [
            {"from": "externalId", "to": "externalId", "asType": "STRING"},
        ]
        select_rows = ["cast(`externalId` as STRING) as externalId"]
        for prop_name, prop in properties.items():
            container = container_by_id.get(prop.container)
            if container is not None:
                dms_type = container.properties[prop.container_property_identifier].type._type
                if dms_type in _DATA_TYPE_BY_DMS_TYPE:
                    sql_type = _DATA_TYPE_BY_DMS_TYPE[dms_type].sql
                else:
                    warnings.warn(
                        f"Unknown DMS type '{dms_type}' for property '{prop_name}'", RuntimeWarning, stacklevel=2
                    )
                    sql_type = "STRING"
            else:
                sql_type = "STRING"
            select_rows.append(f"cast(`{prop_name}` as {sql_type}) as {prop_name}")
            mappings.append({"from": prop_name, "to": prop_name, "asType": sql_type})
        mapping_mode["mappings"] = mappings
        select = ",\n  ".join(select_rows)

        return TransformationWrite(
            external_id=f"{table.name}Transformation",
            name=f"{table.name}Transformation",
            ignore_null_fields=True,
            destination=Nodes(
                view=ViewInfo(view.space, view.external_id, view.version),
                instance_space=instance_space,
            ),
            conflict_mode="upsert",
            query=f"""/* MAPPING_MODE_ENABLED: true */
/* {json.dumps(mapping_mode)} */
select
  {select}
from
  `{table.database}`.`{table.name}`;
""",
        )

    @classmethod
    def _create_edge_transformation(
        cls, property_: dm.EdgeConnectionApply, view: ViewApply, table: RawTableWrite, instance_space: str
    ) -> TransformationWrite:
        start, end = view.external_id, property_.source.external_id
        if property_.direction == "inwards":
            start, end = end, start
        mapping_mode = {
            "version": 1,
            "sourceType": "raw",
            "mappings": [
                {"from": "externalId", "to": "externalId", "asType": "STRING"},
                {"from": start, "to": "startNode", "asType": "STRUCT<`space`:STRING, `externalId`:STRING>"},
                {"from": end, "to": "endNode", "asType": "STRUCT<`space`:STRING, `externalId`:STRING>"},
            ],
            "sourceLevel1": table.database,
            "sourceLevel2": table.name,
        }
        select_rows = [
            "cast(`externalId` as STRING) as externalId",
            f"node_reference('{instance_space}', `{start}`) as startNode",
            f"node_reference('{instance_space}', `{end}`) as endNode",
        ]
        select = ",\n  ".join(select_rows)

        return TransformationWrite(
            external_id=f"{table.name}Transformation",
            name=f"{table.name}Transformation",
            ignore_null_fields=True,
            destination=Edges(
                instance_space=instance_space,
                edge_type=EdgeType(space=property_.type.space, external_id=property_.type.external_id),
            ),
            conflict_mode="upsert",
            query=f"""/* MAPPING_MODE_ENABLED: true */
/* {json.dumps(mapping_mode)} */
select
  {select}
from
  `{table.database}`.`{table.name}`;
""",
        )



"""This module performs importing of graph to TransformationRules pydantic class.
In more details, it traverses the graph and abstracts class and properties, basically
generating a list of rules based on which nodes that form the graph are made.
"""

from pathlib import Path

import pandas as pd
from openpyxl import Workbook, load_workbook

from cognite.neat.utils.auxiliary import local_import

from ._base import BaseImporter


class ExcelImporter(BaseImporter):
    def __init__(self, filepath: Path):
        self.filepath = filepath

    def to_tables(self) -> dict[str, pd.DataFrame]:
        workbook: Workbook = load_workbook(self.filepath)

        return {
            sheet_name: pd.read_excel(
                self.filepath,
                sheet_name=sheet_name,
                header=None if sheet_name == "Metadata" else 0,
                skiprows=1 if sheet_name in ["Classes", "Properties", "Instances"] else None,
            )
            for sheet_name in workbook.sheetnames
        }


class GoogleSheetImporter(BaseImporter):
    def __init__(self, sheet_id: str):
        self.sheet_id = sheet_id

    def to_tables(self) -> dict[str, pd.DataFrame]:
        local_import("gspread", "google")
        import gspread  # type: ignore[import]

        client_google = gspread.service_account()
        google_sheet = client_google.open_by_key(self.sheet_id)

        return {worksheet.title: pd.DataFrame(worksheet.get_all_records()) for worksheet in google_sheet.worksheets()}



from cognite.neat.utils.exceptions import NeatError


class InvalidStepOutputException(NeatError):
    type_ = "invalid_step_output"
    code = 400
    description = "The step output is invalid."
    example = "The step output must be a dictionary."

    def __init__(self, step_type: str):
        self.message = f"Object type {step_type} is not supported as step output"
        super().__init__(self.message)


class ConfigurationNotSet(NeatError):
    type_ = "configuration_not_set"
    code = 400
    description = "The configuration is not set."
    example = "The configuration must be set before running the workflow."

    def __init__(self, config_variable: str):
        self.message = (
            f"The configuration variable '{config_variable}' is not set. "
            f"Please set the configuration before running the workflow."
        )
        super().__init__(self.message)

    def __str__(self):
        return self.message


class StepNotInitialized(NeatError):
    def __init__(self, step_name: str):
        self.message = f"Step {step_name} has not been initialized."
        super().__init__(self.message)


class StepFlowContextNotInitialized(NeatError):
    def __init__(self, step_name: str):
        self.message = f"Step {step_name} requires flow context which is missing."
        super().__init__(self.message)



from typing import cast

import numpy as np
import pandas as pd
from rdflib import OWL, Graph

from cognite.neat.rules.models._base import MatchType
from cognite.neat.utils.utils import remove_namespace


def parse_owl_classes(graph: Graph, make_compliant: bool = False, language: str = "en") -> list[dict]:
    """Parse owl classes from graph to pandas dataframe.

    Args:
        graph: Graph containing owl classes
        make_compliant: Flag for generating compliant classes, by default False
        language: Language to use for parsing, by default "en"

    Returns:
        Dataframe containing owl classes

    !!! note "make_compliant"
        If `make_compliant` is set to True, in presence of errors, default values will be used instead.
        This makes the method very opinionated, but results in a compliant classes.
    """

    query = """
        SELECT ?class ?name ?description ?parentClass ?reference ?match ?comment
        WHERE {
        ?class a owl:Class .
        OPTIONAL {?class rdfs:subClassOf ?parentClass }.
        OPTIONAL {?class rdfs:label ?name }.
        OPTIONAL {?class rdfs:comment ?description} .
        FILTER (!isBlank(?class))
        FILTER (!bound(?parentClass) || !isBlank(?parentClass))
        FILTER (!bound(?name) || LANG(?name) = "" || LANGMATCHES(LANG(?name), "en"))
        FILTER (!bound(?description) || LANG(?description) = "" || LANGMATCHES(LANG(?description), "en"))
    }
    """

    # create raw dataframe

    raw_df = _parse_raw_dataframe(cast(list[tuple], list(graph.query(query.replace("en", language)))))
    if raw_df.empty:
        return []

    # group values and clean up
    processed_df = _clean_up_classes(raw_df)

    # make compliant
    if make_compliant:
        processed_df = make_classes_compliant(processed_df)

    # Make Parent Class list elements into string joined with comma
    processed_df["Parent Class"] = processed_df["Parent Class"].apply(
        lambda x: ", ".join(x) if isinstance(x, list) and x else None
    )

    return processed_df.dropna(axis=0, how="all").replace(float("nan"), None).to_dict(orient="records")


def _parse_raw_dataframe(query_results: list[tuple]) -> pd.DataFrame:
    df = pd.DataFrame(
        query_results,
        columns=["Class", "Name", "Description", "Parent Class", "Reference", "Match", "Comment"],
    )
    if df.empty:
        return df

    # # remove NaNs
    df.replace(np.nan, "", regex=True, inplace=True)

    df.Reference = df.Class
    df.Class = df.Class.apply(lambda x: remove_namespace(x))
    df["Match Type"] = len(df) * [MatchType.exact]
    df["Comment"] = len(df) * [None]
    df["Parent Class"] = df["Parent Class"].apply(lambda x: remove_namespace(x))

    return df


def _clean_up_classes(df: pd.DataFrame) -> pd.DataFrame:
    clean_list = [
        {
            "Class": class_,
            "Name": group_df["Name"].unique()[0],
            "Description": "\n".join(list(group_df.Description.unique())),
            "Parent Class": ", ".join(list(group_df["Parent Class"].unique())),
            "Reference": group_df["Reference"].unique()[0],
            "Match Type": group_df["Match Type"].unique()[0],
            "Comment": group_df["Comment"].unique()[0],
        }
        for class_, group_df in df.groupby("Class")
    ]

    df = pd.DataFrame(clean_list)

    # bring NaNs back
    df.replace("", None, inplace=True)

    # split Parent Class column back into list
    df["Parent Class"] = df["Parent Class"].apply(lambda x: x.split(", ") if isinstance(x, str) else None)

    return df


def make_classes_compliant(classes: pd.DataFrame) -> pd.DataFrame:
    """Make classes compliant.

    Returns:
        Dataframe containing compliant classes

    !!! note "About the compliant classes"
        The compliant classes are based on the OWL base ontology, but adapted to NEAT and use in CDF.
        One thing to note is that this method would not be able to fix issues with class ids which
        are not compliant with the CDF naming convention. For example, if a class id contains a space,
        starts with a number, etc. This will cause issues when trying to create the class in CDF.
    """

    # Replace empty or non-string values in "Match" column with "exact"
    classes["Match Type"] = classes["Match Type"].fillna(MatchType.exact)
    classes["Match Type"] = classes["Match Type"].apply(
        lambda x: MatchType.exact if not isinstance(x, str) or len(x) == 0 else x
    )

    # Replace empty or non-string values in "Comment" column with a default value
    classes["Comment"] = classes["Comment"].fillna("Imported from Ontology by NEAT")
    classes["Comment"] = classes["Comment"].apply(
        lambda x: "Imported from Ontology by NEAT" if not isinstance(x, str) or len(x) == 0 else x
    )

    # Add _object_property_class, _data_type_property_class, _thing_class to the dataframe
    classes = pd.concat(
        [classes, pd.DataFrame([_object_property_class(), _data_type_property_class(), _thing_class()])],
        ignore_index=True,
    )

    # Reduce length of elements in the "Description" column to 1024 characters
    classes["Description"] = classes["Description"].apply(lambda x: x[:1024] if isinstance(x, str) else None)

    # Add missing parent classes to the dataframe
    classes = pd.concat(
        [classes, pd.DataFrame(_add_parent_class(classes))],
        ignore_index=True,
    )

    return classes


def _object_property_class() -> dict:
    return {
        "Class": "ObjectProperty",
        "Name": None,
        "Description": "The class of object properties.",
        "Parent Class": None,
        "Reference": OWL.ObjectProperty,
        "Match Type": MatchType.exact,
        "Comment": "Added by NEAT based on owl:ObjectProperty but adapted to NEAT and use in CDF.",
    }


def _data_type_property_class() -> dict:
    return {
        "Class": "DatatypeProperty",
        "Name": None,
        "Description": "The class of data properties.",
        "Parent Class": None,
        "Reference": OWL.DatatypeProperty,
        "Match Type": MatchType.exact,
        "Comment": "Added by NEAT based on owl:DatatypeProperty but adapted to NEAT and use in CDF.",
    }


def _thing_class() -> dict:
    return {
        "Class": "Thing",
        "Name": None,
        "Description": "The class of holding class individuals.",
        "Parent Class": None,
        "Reference": OWL.Thing,
        "Match Type": MatchType.exact,
        "Comment": (
            "Added by NEAT. "
            "Imported from OWL base ontology, it is meant for use as a default"
            " value type for object properties which miss a declared range."
        ),
    }


def _add_parent_class(df: pd.DataFrame) -> list[dict]:
    parent_set = {
        item
        for sublist in df["Parent Class"].tolist()
        if sublist
        for item in sublist
        if item != "" and item is not None
    }
    class_set = set(df["Class"].tolist())

    rows = []
    for missing_parent_class in parent_set.difference(class_set):
        rows += [
            {
                "Class": missing_parent_class,
                "Name": None,
                "Description": None,
                "Parent Class": None,
                "Reference": None,
                "Match Type": None,
                "Comment": (
                    "Added by NEAT. "
                    "This is a parent class that is missing in the ontology. "
                    "It is added by NEAT to make the ontology compliant with CDF."
                ),
            }
        ]

    return rows



import re
import xml.etree.ElementTree as ET
from pathlib import Path
from xml.etree.ElementTree import Element

from rdflib import OWL, RDF, RDFS, SKOS, XSD, Literal, Namespace, URIRef

from cognite.neat.constants import PREFIXES
from cognite.neat.legacy.graph.models import Triple
from cognite.neat.legacy.rules.models._base import ALLOWED_PATTERN
from cognite.neat.utils.utils import get_namespace, remove_namespace
from cognite.neat.utils.xml import get_children, iterate_tree

from ._base import BaseExtractor

_DEXPI_PREFIXES = {
    "dexpi": Namespace("http://sandbox.dexpi.org/rdl/"),
    "posccaesar": Namespace("http://data.posccaesar.org/rdl/"),
}


class DexpiXML(BaseExtractor):
    """
    DEXPI-XML extractor of RDF triples

    Args:
        filepath: File path to DEXPI XML file.
        namespace: Optional custom namespace to use for extracted triples that define data
                    model instances. Defaults to http://purl.org/cognite/neat/.
    """

    def __init__(
        self,
        filepath: Path | str,
        base_namespace: str | None = None,
    ):
        self.filepath = Path(filepath)
        self.namespace = Namespace(base_namespace) if isinstance(base_namespace, str | Namespace) else PREFIXES["neat"]

    def extract(self) -> set[Triple]:
        """
        Extracts RDF triples from the graph capturing sheet.

        Returns:
            List of RDF triples, represented as tuples `(subject, predicate, object)`, that define data model instances
        """
        if self.filepath is None:
            raise ValueError("File path to the graph capturing sheet is not provided!")

        root = ET.parse(self.filepath).getroot()

        nodes, edges = _extract_nodes_and_edges(root)

        self.nodes = nodes

        return _to_triples(nodes, edges, base_namespace=self.namespace)


def _extract_nodes_and_edges(root: Element) -> tuple[dict, dict]:
    """Extract nodes and edges from an XML tree.

    Args:
        root: XML tree to extract nodes and edges from.

    Returns:
        Tuple of nodes and edges.
    """
    nodes: dict = {}
    edges: dict = {"associations": set(), "connections": set(), "children": set()}

    for element in iterate_tree(root):
        if (
            "ComponentClass" in element.attrib
            and element.attrib["ComponentClass"] != "Label"
            and "ID" in element.attrib
        ):
            id_ = element.attrib["ID"]

            # add header that contains things such as type
            nodes[id_] = {"header": element.attrib, "attributes": {}}
            nodes[id_]["header"]["tag"] = str(element.tag)

            # add human readable label of node (if exists)
            nodes = _add_node_label(id_, nodes, element)

            # add generic attributes of node
            nodes = _add_node_generic_attributes(id_, nodes, element)

            # edges scenario: <Connection> tag
            edges = _add_connection_edge(edges, element)

            # edges scenario: <Association> tag
            edges = _add_association_edge(id_, edges, element)

            # edge scenario: children elements of the element
            edges = _add_child_edge(id_, edges, element)

    return nodes, edges


def _add_node_label(id_: str, nodes: dict, element: Element) -> dict:
    if children := get_children(element, "Label", 1):
        if grandchildren := get_children(children[0], "Text", 1):
            if "String" in grandchildren[0].attrib:
                nodes[id_]["header"]["label"] = grandchildren[0].attrib["String"]
    # extension for schema version 3.3, where text is used to "label" without a <label> parent
    elif children := get_children(element, "Text", 1):
        if "String" in children[0].attrib:
            nodes[id_]["header"]["label"] = children[0].attrib["String"]

    return nodes


def _add_node_generic_attributes(id_: str, nodes: dict, element: Element) -> dict:
    if children := get_children(element, "GenericAttributes", 1):
        if grandchildren := get_children(children[0], "GenericAttribute"):
            for generic_attribute in grandchildren:
                # extension for schema version 3.3, where "AttributeURI" is not included
                if "AttributeURI" in generic_attribute.attrib:
                    if generic_attribute.attrib["AttributeURI"] not in nodes[id_]:
                        nodes[id_]["attributes"][generic_attribute.attrib["AttributeURI"]] = [generic_attribute.attrib]
                    else:
                        nodes[id_]["attributes"][generic_attribute.attrib["AttributeURI"]].append(
                            generic_attribute.attrib
                        )

    return nodes


def _add_connection_edge(edges: dict, element: Element) -> dict:
    if "connections" not in edges:
        edges["connections"] = set()

    if connections := get_children(element, "Connection"):
        for connection in connections:
            if "FromID" in connection.attrib and "ToID" in connection.attrib:
                edges["connections"].add((connection.attrib["FromID"], "connection", connection.attrib["ToID"]))

    return edges


def _add_association_edge(id_: str, edges: dict, element: Element) -> dict:
    if "associations" not in edges:
        edges["associations"] = set()

    if associations := get_children(element, "Association"):
        for association in associations:
            if "Type" in association.attrib and "ItemID" in association.attrib:
                association_type = "".join(
                    [
                        word.capitalize() if i != 0 else word
                        for i, word in enumerate(association.attrib["Type"].split(" "))
                    ]
                )
                edges["associations"].add((id_, f"{association_type}", association.attrib["ItemID"]))

    return edges


def _add_child_edge(id_: str, edges: dict, element: Element) -> dict:
    if "children" not in edges:
        edges["children"] = set()

    for child in element:
        if "ID" in child.attrib and child.tag != "Label":
            edges["children"].add((id_, child.tag, child.attrib["ID"]))

    return edges


def _to_triples(nodes: dict, edges: dict, base_namespace: Namespace) -> set[Triple]:
    """Convert nodes and edges to subject-predicate-object triples.

    Args:
        nodes: Nodes to convert to triples.
        edges: Edges to convert to triples.

    Returns:
        List of triples.
    """
    triples: set[Triple] = set()

    # Adding nodes and nodes attributes
    for id_, node in nodes.items():
        uri = URIRef(base_namespace + id_)

        # add header attribute triples
        triples = _add_header_triples(uri, triples, node["header"], "ComponentClassURI", RDF.type)
        triples = _add_header_triples(uri, triples, node["header"], "label", RDFS.label)
        triples = _add_header_triples(uri, triples, node["header"], "tag")
        triples = _add_header_triples(uri, triples, node["header"], "ComponentClass")
        triples = _add_header_triples(uri, triples, node["header"], "ComponentName")

        # add generic attribute triples
        for attribute, values in node["attributes"].items():
            attribute_uri, triples = _to_compliant_attribute_uri(URIRef(attribute), triples)

            for value in values:
                triples = _add_attribute_triples(uri, triples, attribute_uri, value)

    # Adding Edges
    triples = _add_edge_triples(edges["connections"], triples, "connections", base_namespace)
    triples = _add_edge_triples(edges["associations"], triples, "associations", base_namespace)
    triples = _add_edge_triples(edges["children"], triples, "children", base_namespace)

    return triples


def _add_edge_triples(
    edges: set[tuple[str, str, str]], triples: set[Triple], edge_type: str, base_namespace: Namespace
) -> set[Triple]:
    """Convert edges to subject-predicate-object triples.

    Args:
        edges: Edges to convert to triples.

    Returns:
        List of triples.
    """

    for subject, predicate, object in edges:
        if edge_type != "connections":
            predicate_uri = _DEXPI_PREFIXES["dexpi"][f"{edge_type}/{predicate}"]
        else:
            predicate_uri = _DEXPI_PREFIXES["dexpi"][predicate]

        triples.add((URIRef(base_namespace + subject), predicate_uri, URIRef(base_namespace + object)))

    return triples


def _add_header_triples(
    uri: URIRef, triples: set[Triple], attributes: dict, attribute: str, attribute_uri: URIRef | None = None
) -> set[Triple]:
    if attribute not in attributes:
        return triples

    attribute_uri = attribute_uri or _DEXPI_PREFIXES["dexpi"][attribute]

    if attribute in attributes and attribute_uri == RDF.type:
        triples.add((uri, attribute_uri, URIRef(attributes[attribute])))
    else:
        triples.add((uri, attribute_uri, Literal(attributes[attribute])))

    return triples


def _to_compliant_attribute_uri(attribute_uri: URIRef, triples: set[Triple]) -> tuple[URIRef, set[Triple]]:
    namespace = get_namespace(attribute_uri)
    id_ = remove_namespace(attribute_uri)
    compliant_id = re.sub(ALLOWED_PATTERN, "", id_)
    compliant_attribute_uri = URIRef(f"{namespace}{compliant_id}")

    if attribute_uri != compliant_attribute_uri:
        triples.add(
            (
                compliant_attribute_uri,
                SKOS.exactMatch,
                attribute_uri,
            )
        )
        triples.add(
            (
                compliant_attribute_uri,
                RDFS.comment,
                Literal("Modified property URI to be compliant with NEAT internal representation"),
            )
        )

    return compliant_attribute_uri, triples


def _add_attribute_triples(uri: URIRef, triples: set[Triple], attribute_uri, value: dict) -> set[Triple]:
    # case: when Value or Format is not present we skip the attribute since it is not formatted correctly
    if "Value" not in value or "Format" not in value:
        return triples

    # case: when unit is present we create a special datatype for it
    if "Units" in value:
        triples.update(_generate_datatype_triples(_DEXPI_PREFIXES["dexpi"], value))
        triples.add(
            (
                uri,
                attribute_uri,
                Literal(value["Value"], datatype=_DEXPI_PREFIXES["dexpi"][value["Units"]]),
            )
        )

    # case: when language is present we create add language tag to the literal
    elif "Language" in value and "Value" in value:
        triples.add((uri, attribute_uri, Literal(value["Value"], lang=value["Language"])))

    # case: when ValueURI is present we use it instead of Value
    elif "ValueURI" in value:
        triples.add((uri, attribute_uri, Literal(value["ValueURI"], datatype=XSD[value["Format"]])))

    # case: when Format is not string we make sure to add the datatype
    elif value["Format"].lower() != "string":
        triples.add((uri, attribute_uri, Literal(value["Value"], datatype=XSD[value["Format"]])))

    # case: when Format is string we add the literal without datatype (easier to read triples, less noise)
    else:
        triples.add((uri, attribute_uri, Literal(value["Value"])))

    return triples


def _generate_datatype_triples(base_namespace: Namespace, unit_definition: dict) -> set[Triple]:
    return {
        (base_namespace[unit_definition["Units"]], RDF.type, RDFS.Datatype),
        (base_namespace[unit_definition["Units"]], RDFS.label, Literal(unit_definition["Units"])),
        (base_namespace[unit_definition["Units"]], OWL.equivalentClass, XSD[unit_definition["Format"]]),
        (base_namespace[unit_definition["Units"]], SKOS.exactMatch, URIRef(unit_definition["UnitsURI"])),
    }



import time
from pathlib import Path
from typing import ClassVar, Literal, cast

from cognite.neat.rules import exporters
from cognite.neat.rules._shared import DMSRules, InformationRules, Rules
from cognite.neat.rules.models import RoleTypes
from cognite.neat.workflows._exceptions import StepNotInitialized
from cognite.neat.workflows.model import FlowMessage, StepExecutionStatus
from cognite.neat.workflows.steps.data_contracts import CogniteClient, MultiRuleData
from cognite.neat.workflows.steps.step_model import Configurable, Step

__all__ = [
    "RulesToDMS",
    "RulesToExcel",
    "RulesToOntology",
    "RulesToSHACL",
    "RulesToSemanticDataModel",
    "RulesToCDFTransformations",
    "DeleteDataModelFromCDF",
]


CATEGORY = __name__.split(".")[-1].replace("_", " ").title()


class DeleteDataModelFromCDF(Step):
    """
    This step deletes data model and its components from CDF
    """

    description = "This step deletes data model and its components from CDF."
    version = "private-beta"
    category = CATEGORY

    configurables: ClassVar[list[Configurable]] = [
        Configurable(
            name="Dry run",
            value="False",
            label=("Whether to perform a dry run of the deleter. "),
            options=["True", "False"],
        ),
        Configurable(
            name="Components",
            type="multi_select",
            value="",
            label="Select which DMS schema component(s) to be deleted from CDF",
            options=["spaces", "containers", "views", "data_models"],
        ),
        Configurable(
            name="Multi-space components deletion",
            value="False",
            label=(
                "Whether to delete only components belonging to the data model space"
                " (i.e. space define under Metadata sheet of Rules), "
                "or also additionally delete components outside of the data model space."
            ),
            options=["True", "False"],
        ),
    ]

    def run(self, rules: MultiRuleData, cdf_client: CogniteClient) -> FlowMessage:  # type: ignore[override]
        if self.configs is None or self.data_store_path is None:
            raise StepNotInitialized(type(self).__name__)
        components_to_delete = {
            cast(Literal["all", "spaces", "data_models", "views", "containers"], key)
            for key, value in self.complex_configs["Components"].items()
            if value
        }
        dry_run = self.configs["Dry run"] == "True"
        multi_space_components_delete: bool = self.configs["Multi-space components deletion"] == "True"

        if not components_to_delete:
            return FlowMessage(
                error_text="No DMS Schema components selected for removal! Please select minimum one!",
                step_execution_status=StepExecutionStatus.ABORT_AND_FAIL,
            )
        input_rules = rules.dms or rules.information
        if input_rules is None:
            return FlowMessage(
                error_text="Missing DMS or Information rules in the input data! "
                "Please ensure that a DMS or Information rules is provided!",
                step_execution_status=StepExecutionStatus.ABORT_AND_FAIL,
            )

        dms_exporter = exporters.DMSExporter(
            export_components=frozenset(components_to_delete),
            include_space=(
                None
                if multi_space_components_delete
                else {input_rules.metadata.space if isinstance(input_rules, DMSRules) else input_rules.metadata.prefix}
            ),
        )

        report_lines = ["# Data Model Deletion from CDF\n\n"]
        errors = []
        for result in dms_exporter.delete_from_cdf(rules=input_rules, client=cdf_client, dry_run=dry_run):
            report_lines.append(result.as_report_str())
            errors.extend(result.error_messages)

        report_lines.append("\n\n# ERRORS\n\n")
        report_lines.extend(errors)

        output_dir = self.config.staging_path
        output_dir.mkdir(parents=True, exist_ok=True)
        report_file = "dms_component_creation_report.txt"
        report_full_path = output_dir / report_file
        report_full_path.write_text("\n".join(report_lines))

        output_text = (
            "<p></p>"
            "Download Data Model Deletion "
            f'<a href="/data/staging/{report_file}?{time.time()}" '
            f'target="_blank">Report</a>'
        )

        if errors:
            return FlowMessage(error_text=output_text, step_execution_status=StepExecutionStatus.ABORT_AND_FAIL)
        else:
            return FlowMessage(output_text=output_text)


class RulesToDMS(Step):
    """
    This step exports Rules to DMS Schema components in CDF
    """

    description = "This step exports Rules to DMS Schema components in CDF."
    version = "private-beta"
    category = CATEGORY

    configurables: ClassVar[list[Configurable]] = [
        Configurable(
            name="Dry run",
            value="False",
            label=("Whether to perform a dry run of the export. "),
            options=["True", "False"],
        ),
        Configurable(
            name="Components",
            type="multi_select",
            value="",
            label="Select which DMS schema component(s) to export to CDF",
            options=["spaces", "containers", "views", "data_models"],
        ),
        Configurable(
            name="Existing component handling",
            value="fail",
            label=(
                "How to handle situation when components being exported in CDF already exist."
                "Fail the step if any component already exists, "
                "Skip the component if it already exists, "
                " or Update the component try to update the component."
            ),
            options=["fail", "skip", "update", "force"],
        ),
        Configurable(
            name="Multi-space components create",
            value="False",
            label=(
                "Whether to create only components belonging to the data model space"
                " (i.e. space define under Metadata sheet of Rules), "
                "or also additionally components outside of the data model space."
            ),
            options=["True", "False"],
        ),
    ]

    def run(self, rules: MultiRuleData, cdf_client: CogniteClient) -> FlowMessage:  # type: ignore[override]
        if self.configs is None or self.data_store_path is None:
            raise StepNotInitialized(type(self).__name__)
        existing_components_handling = cast(
            Literal["fail", "update", "skip", "force"], self.configs["Existing component handling"]
        )
        multi_space_components_create: bool = self.configs["Multi-space components create"] == "True"
        components_to_create = {
            cast(Literal["all", "spaces", "data_models", "views", "containers"], key)
            for key, value in self.complex_configs["Components"].items()
            if value
        }
        dry_run = self.configs["Dry run"] == "True"

        if not components_to_create:
            return FlowMessage(
                error_text="No DMS Schema components selected for upload! Please select minimum one!",
                step_execution_status=StepExecutionStatus.ABORT_AND_FAIL,
            )
        input_rules = rules.dms or rules.information
        if input_rules is None:
            return FlowMessage(
                error_text="Missing DMS or Information rules in the input data! "
                "Please ensure that a DMS or Information rules is provided!",
                step_execution_status=StepExecutionStatus.ABORT_AND_FAIL,
            )

        dms_exporter = exporters.DMSExporter(
            export_components=frozenset(components_to_create),
            include_space=(
                None
                if multi_space_components_create
                else {input_rules.metadata.space if isinstance(input_rules, DMSRules) else input_rules.metadata.prefix}
            ),
            existing_handling=existing_components_handling,
        )

        output_dir = self.config.staging_path
        output_dir.mkdir(parents=True, exist_ok=True)
        file_name = (
            input_rules.metadata.external_id
            if isinstance(input_rules, DMSRules)
            else input_rules.metadata.name.replace(" ", "_").lower()
        )
        schema_zip = f"{file_name}.zip"
        schema_full_path = output_dir / schema_zip
        dms_exporter.export_to_file(input_rules, schema_full_path)

        report_lines = ["# DMS Schema Export to CDF\n\n"]
        errors = []
        for result in dms_exporter.export_to_cdf(rules=input_rules, client=cdf_client, dry_run=dry_run):
            report_lines.append(result.as_report_str())
            errors.extend(result.error_messages)

        report_lines.append("\n\n# ERRORS\n\n")
        report_lines.extend(errors)

        output_dir = self.config.staging_path
        output_dir.mkdir(parents=True, exist_ok=True)
        report_file = "dms_component_creation_report.txt"
        report_full_path = output_dir / report_file
        report_full_path.write_text("\n".join(report_lines))

        output_text = (
            "<p></p>"
            "Download DMS Export "
            f'<a href="/data/staging/{report_file}?{time.time()}" '
            f'target="_blank">Report</a>'
            "<p></p>"
            "Download DMS exported schema"
            f'- <a href="/data/staging/{schema_zip}?{time.time()}" '
            f'target="_blank">{schema_zip}</a>'
        )

        if errors:
            return FlowMessage(error_text=output_text, step_execution_status=StepExecutionStatus.ABORT_AND_FAIL)
        else:
            return FlowMessage(output_text=output_text)


class RulesToExcel(Step):
    """This step exports Rules to Excel serialization"""

    description = "This step exports Rules to Excel serialization."
    version = "private-beta"
    category = CATEGORY

    configurables: ClassVar[list[Configurable]] = [
        Configurable(
            name="Styling",
            value="default",
            label="Styling of the Excel file",
            options=list(exporters.ExcelExporter.style_options),
        ),
        Configurable(
            name="Output role format",
            value="input",
            label="The role to use for the exported spreadsheet. If provided, the rules will be converted to "
            "this role format before being written to excel. If not provided, the role from the input "
            "rules will be used.",
            options=["input", *RoleTypes.__members__.keys()],
        ),
        Configurable(
            name="Dump Format",
            value="user",
            label="How to dump the rules to the Excel file.\n"
            "'user' - just as is.\n'reference' - enterprise model used as basis for a solution model\n"
            "'last' - used when updating a data model.",
            options=list(exporters.ExcelExporter.dump_options),
        ),
        Configurable(
            name="New Data Model ID",
            value="",
            label="If you chose Dump Format 'reference', the provided ID will be use in the new medata sheet. "
            "Expected format 'sp_space:my_external_id'.",
            options=list(exporters.ExcelExporter.dump_options),
        ),
        Configurable(
            name="File path",
            value="",
            label="File path to the generated Excel file.For example: 'staging/exported-rules.xlsx'",
        ),
    ]

    def run(self, rules: MultiRuleData) -> FlowMessage:  # type: ignore[override, syntax]
        if self.configs is None or self.data_store_path is None:
            raise StepNotInitialized(type(self).__name__)

        dump_format = self.configs.get("Dump Format", "user")
        styling = cast(exporters.ExcelExporter.Style, self.configs.get("Styling", "default"))
        role = self.configs.get("Output role format")
        output_role = None
        if role != "input" and role is not None:
            output_role = RoleTypes[role]

        new_model_str = self.configs.get("New Data Model ID")
        new_model_id: tuple[str, str] | None = None
        if new_model_str and ":" in new_model_str:
            new_model_id = tuple(new_model_str.split(":", 1))  # type: ignore[assignment]
        elif new_model_str:
            return FlowMessage(
                error_text="New Data Model ID must be in the format 'sp_space:my_external_id'!",
                step_execution_status=StepExecutionStatus.ABORT_AND_FAIL,
            )

        excel_exporter = exporters.ExcelExporter(
            styling=styling,
            output_role=output_role,
            dump_as=dump_format,  # type: ignore[arg-type]
            new_model_id=new_model_id,
        )

        rule_instance: Rules
        if rules.domain:
            rule_instance = rules.domain
        elif rules.information:
            rule_instance = rules.information
        elif rules.dms:
            rule_instance = rules.dms
        else:
            output_errors = "No rules provided for export!"
            return FlowMessage(error_text=output_errors, step_execution_status=StepExecutionStatus.ABORT_AND_FAIL)
        if output_role is None:
            output_role = rule_instance.metadata.role
        output_dir = self.data_store_path / Path("staging")
        output_dir.mkdir(parents=True, exist_ok=True)
        file_name = f"exported_rules_{output_role.value}.xlsx"
        filepath = output_dir / file_name
        if self.configs.get("File path", ""):
            file_name = self.configs["File path"]
            filepath = Path(self.data_store_path) / Path(file_name)
        else:
            file_name = f"staging/{file_name}"

        excel_exporter.export_to_file(rule_instance, filepath)

        output_text = (
            "<p></p>"
            f"Download Excel Exported {output_role.value} rules: "
            f'- <a href="/data/{file_name}?{time.time()}" '
            f'target="_blank">{file_name}</a>'
        )

        return FlowMessage(output_text=output_text)


class RulesToOntology(Step):
    """
    This step exports Rules to OWL ontology
    """

    description = "This step exports Rules to OWL ontology"
    version = "private-beta"
    category = CATEGORY
    configurables: ClassVar[list[Configurable]] = [
        Configurable(
            name="File path",
            value="staging/ontology.ttl",
            label=("Relative path for the ontology file storage, " " It will be auto-created if not provided ! "),
        )
    ]

    def run(self, rules: MultiRuleData) -> FlowMessage:  # type: ignore[override, syntax]
        if self.configs is None or self.data_store_path is None:
            raise StepNotInitialized(type(self).__name__)

        if not rules.information and not rules.dms:
            return FlowMessage(
                error_text="Rules must be made either by Information Architect or DMS Architect!",
                step_execution_status=StepExecutionStatus.ABORT_AND_FAIL,
            )

        default_path = self.config.staging_path / _get_default_file_name(rules, "ontology", "ttl")

        storage_path = (
            self.data_store_path / Path(self.configs["File path"]) if self.configs["File path"] else default_path
        )
        storage_path.parent.mkdir(parents=True, exist_ok=True)

        exporter = exporters.OWLExporter()
        exporter.export_to_file(cast(InformationRules | DMSRules, rules.information or rules.dms), storage_path)

        relative_file_path = "/".join(storage_path.relative_to(self.data_store_path).parts)

        output_text = (
            "<p></p>"
            "Rules exported as OWL ontology can be downloaded here : "
            f'<a href="/data/{relative_file_path}?{time.time()}" '
            f'target="_blank">{storage_path.stem}.ttl</a>'
        )

        return FlowMessage(output_text=output_text)


class RulesToSHACL(Step):
    """
    This step exports Rules to SHACL
    """

    description = "This step exports Rules to SHACL"
    version = "private-beta"
    category = CATEGORY
    configurables: ClassVar[list[Configurable]] = [
        Configurable(
            name="File path",
            value="staging/shacl.ttl",
            label=(
                "Relative path for the shacl file storage, "
                "must end with .ttl ! It will be auto-created if not provided !"
            ),
        )
    ]

    def run(self, rules: MultiRuleData) -> FlowMessage:  # type: ignore[override, syntax]
        if self.configs is None or self.data_store_path is None:
            raise StepNotInitialized(type(self).__name__)

        if not rules.information and not rules.dms:
            return FlowMessage(
                error_text="Rules must be made either by Information Architect or DMS Architect!",
                step_execution_status=StepExecutionStatus.ABORT_AND_FAIL,
            )

        default_path = self.config.staging_path / _get_default_file_name(rules, "shacl", "ttl")

        storage_path = (
            self.data_store_path / Path(self.configs["File path"]) if self.configs["File path"] else default_path
        )
        storage_path.parent.mkdir(parents=True, exist_ok=True)

        exporter = exporters.SHACLExporter()
        exporter.export_to_file(cast(InformationRules | DMSRules, rules.information or rules.dms), storage_path)

        relative_file_path = "/".join(storage_path.relative_to(self.data_store_path).parts)

        output_text = (
            "<p></p>"
            "Rules exported as SHACL shapes can be downloaded here : "
            f'<a href="/data/{relative_file_path}?{time.time()}" '
            f'target="_blank">{storage_path.stem}.ttl</a>'
        )

        return FlowMessage(output_text=output_text)


class RulesToSemanticDataModel(Step):
    """
    This step exports Rules to semantic data model
    """

    description = "This step exports Rules to semantic data model (ontology + SHACL)"
    version = "private-beta"
    category = CATEGORY
    configurables: ClassVar[list[Configurable]] = [
        Configurable(
            name="File path",
            value="staging/semantic-data-model.ttl",
            label=(
                "Relative path for the semantic data model file storage, "
                "must end with .ttl ! It will be auto-created if not provided !"
            ),
        )
    ]

    def run(self, rules: MultiRuleData) -> FlowMessage:  # type: ignore[override, syntax]
        if self.configs is None or self.data_store_path is None:
            raise StepNotInitialized(type(self).__name__)

        if not rules.information and not rules.dms:
            return FlowMessage(
                error_text="Rules must be made either by Information Architect or DMS Architect!",
                step_execution_status=StepExecutionStatus.ABORT_AND_FAIL,
            )

        default_path = self.config.staging_path / _get_default_file_name(rules, "semantic-data-model", "ttl")

        storage_path = (
            self.data_store_path / Path(self.configs["File path"]) if self.configs["File path"] else default_path
        )
        storage_path.parent.mkdir(parents=True, exist_ok=True)

        exporter = exporters.SemanticDataModelExporter()
        exporter.export_to_file(cast(InformationRules | DMSRules, rules.information or rules.dms), storage_path)

        relative_file_path = "/".join(storage_path.relative_to(self.data_store_path).parts)

        output_text = (
            "<p></p>"
            "Rules exported as semantic data model (OWL + SHACL) can be downloaded here : "
            f'<a href="/data/{relative_file_path}?{time.time()}" '
            f'target="_blank">{storage_path.stem}.ttl</a>'
        )

        return FlowMessage(output_text=output_text)


class RulesToCDFTransformations(Step):
    description = "This step exports transformations and RAW tables to populate a data model in CDF"
    version = "private-alpha"
    category = CATEGORY
    configurables: ClassVar[list[Configurable]] = [
        Configurable(
            name="Dry run",
            value="False",
            label=("Whether to perform a dry run of the export. "),
            options=["True", "False"],
        ),
        Configurable(
            name="Instance space",
            value="",
            label=(
                "The space to use for the transformations instances. If provided, "
                "the transformations will be set to populate"
                "this space. If not provided, the space from the input rules will be used."
            ),
        ),
    ]

    def run(self, rules: MultiRuleData, cdf_client: CogniteClient) -> FlowMessage:  # type: ignore[override]
        if self.configs is None or self.data_store_path is None:
            raise StepNotInitialized(type(self).__name__)

        input_rules = rules.dms or rules.information
        if input_rules is None:
            return FlowMessage(
                error_text="Missing DMS or Information rules in the input data! "
                "Please ensure that a DMS or Information rules is provided!",
                step_execution_status=StepExecutionStatus.ABORT_AND_FAIL,
            )
        default_instance_space = (
            input_rules.metadata.space if isinstance(input_rules, DMSRules) else input_rules.metadata.prefix
        )
        instance_space = self.configs.get("Instance space") or default_instance_space
        dry_run = self.configs.get("Dry run", "False") == "True"
        dms_exporter = exporters.DMSExporter(
            export_pipeline=True, instance_space=instance_space, export_components=["spaces"]
        )
        output_dir = self.config.staging_path
        output_dir.mkdir(parents=True, exist_ok=True)
        file_name = (
            input_rules.metadata.external_id
            if isinstance(input_rules, DMSRules)
            else input_rules.metadata.name.replace(" ", "_").lower()
        )
        schema_zip = f"{file_name}_pipeline.zip"
        schema_full_path = output_dir / schema_zip
        dms_exporter.export_to_file(input_rules, schema_full_path)

        report_lines = ["# DMS Schema Export to CDF\n\n"]
        errors = []
        for result in dms_exporter.export_to_cdf(rules=input_rules, client=cdf_client, dry_run=dry_run):
            report_lines.append(result.as_report_str())
            errors.extend(result.error_messages)

        report_lines.append("\n\n# ERRORS\n\n")
        report_lines.extend(errors)

        output_dir = self.config.staging_path
        output_dir.mkdir(parents=True, exist_ok=True)
        report_file = "pipeline_creation_report.txt"
        report_full_path = output_dir / report_file
        report_full_path.write_text("\n".join(report_lines))

        output_text = (
            "<p></p>"
            "Download Pipeline Export "
            f'<a href="/data/staging/{report_file}?{time.time()}" '
            f'target="_blank">Report</a>'
            "<p></p>"
            "Download Pipeline exported schema"
            f'- <a href="/data/staging/{schema_zip}?{time.time()}" '
            f'target="_blank">{schema_zip}</a>'
        )

        if errors:
            return FlowMessage(error_text=output_text, step_execution_status=StepExecutionStatus.ABORT_AND_FAIL)
        else:
            return FlowMessage(output_text=output_text)


def _get_default_file_name(rules: MultiRuleData, file_category: str = "ontology", extension: str = "ttl") -> str:
    name = rules.information.metadata.prefix if rules.information else cast(DMSRules, rules.dms).metadata.space
    version = rules.information.metadata.version if rules.information else cast(DMSRules, rules.dms).metadata.version
    return f"{name}-v{version.strip().replace('.', '_')}-{file_category}.{extension}"



import importlib
from types import ModuleType

from cognite.neat.exceptions import NeatImportError


def local_import(module: str, extra: str) -> ModuleType:
    try:
        return importlib.import_module(module)
    except ImportError as e:
        raise NeatImportError(module.split(".")[0], extra) from e



from .graph_extractor import *  # noqa
from .graph_transformer import *  # noqa
from .graph_loader import *  # noqa
from .graph_store import *  # noqa
from .rules_importer import *  # noqa
from .rules_exporter import *  # noqa
from .graph_contextualization import *  # noqa



from ._rules import DMSContainer, DMSMetadata, DMSProperty, DMSRules, DMSView
from ._rules_input import DMSContainerInput, DMSMetadataInput, DMSPropertyInput, DMSRulesInput, DMSViewInput
from ._schema import DMSSchema, PipelineSchema

__all__ = [
    "DMSRules",
    "DMSSchema",
    "DMSMetadata",
    "DMSView",
    "DMSProperty",
    "DMSContainer",
    "PipelineSchema",
    "DMSRulesInput",
    "DMSMetadataInput",
    "DMSViewInput",
    "DMSPropertyInput",
    "DMSContainerInput",
]



import importlib
import inspect
import logging
import os
import sys
import types
from pathlib import Path
from typing import Any

from pydantic import BaseModel

# steps
import cognite.neat.workflows.steps.lib.current
import cognite.neat.workflows.steps.lib.io
import cognite.neat.workflows.steps.lib.legacy
from cognite.neat.app.monitoring.metrics import NeatMetricsCollector
from cognite.neat.config import Config
from cognite.neat.exceptions import InvalidWorkFlowError
from cognite.neat.workflows._exceptions import ConfigurationNotSet
from cognite.neat.workflows.model import FlowMessage, WorkflowConfigs
from cognite.neat.workflows.steps.step_model import Configurable, DataContract, Step


class StepMetadata(BaseModel):
    name: str
    description: str = ""
    category: str = "default"  # defines the category the step belongs to (e.g. "data", "model", "test")
    scope: str = "global"  # defines the scope of the step (e.g. "global", local to a specific workflow)
    input: list[str]
    output: list[str]
    configurables: list[Configurable] = []
    version: str = "1.0.0"  # version of the step. All alpha versions considered as experimental
    docs_url: str = ""  # url to the extended documentation of the step
    source: str = (
        "cognite"  # source of the step , can be source identifier or url , for instance github url for instance.
    )


class StepsRegistry:
    def __init__(self, config: Config):
        self.config = config
        self._step_classes: list[type[Step]] = []
        self.user_steps_path: Path = config.data_store_path / "steps"
        self.data_store_path: str = str(config.data_store_path)
        self.categorized_steps: dict[str, set] = {"legacy": set(), "current": set(), "io": set()}

    def load_step_classes(self):
        if self._step_classes:
            # classes already loaded - no need to reload
            return
        for name, step_cls in inspect.getmembers(cognite.neat.workflows.steps.lib.current):
            if inspect.isclass(step_cls):
                logging.info(f"Loading NEAT step {name}")
                self._step_classes.append(step_cls)
                self.categorized_steps["current"].add(name)
        for name, step_cls in inspect.getmembers(cognite.neat.workflows.steps.lib.io):
            if inspect.isclass(step_cls):
                logging.info(f"Loading NEAT step {name}")
                self._step_classes.append(step_cls)
                self.categorized_steps["io"].add(name)
        for name, step_cls in inspect.getmembers(cognite.neat.workflows.steps.lib.legacy):
            if inspect.isclass(step_cls):
                logging.info(f"Loading NEAT step {name}")
                self._step_classes.append(step_cls)
                self.categorized_steps["legacy"].add(name)
        sys.path.append(str(Path(self.data_store_path) / "workflows"))
        try:
            if self.user_steps_path:
                sys.path.append(str(self.user_steps_path))
                self.load_custom_step_classes(self.user_steps_path, scope="user_global")
        except Exception as e:
            logging.info(f"No user defined modules provided in {self.user_steps_path}. Error: {e}")

    def load_workflow_step_classes(self, workflow_name: str):
        if not self.data_store_path:
            raise ConfigurationNotSet("data_store_path")
        workflow_steps_path = Path(self.data_store_path) / "workflows" / workflow_name
        if workflow_steps_path.exists():
            self.load_custom_step_classes(workflow_steps_path, scope="workflow")

    def load_custom_step_classes(self, custom_steps_path: Path, scope: str = "global"):
        for step_module_name in os.listdir(custom_steps_path):
            step_module_path = custom_steps_path / Path(step_module_name)
            if step_module_name.startswith("__") or (
                step_module_path.is_file() and not step_module_name.endswith(".py")
            ):
                continue
            if scope == "user_global":
                full_module_name = step_module_name.replace(".py", "")
            else:
                full_module_name = custom_steps_path.name + "." + step_module_name.replace(".py", "")
            logging.info(f"Loading user defined step module {full_module_name}")
            if full_module_name in sys.modules:
                logging.info(f"Reloading existing workflow module {full_module_name}")
                steps_module = importlib.reload(sys.modules[full_module_name])
            else:
                steps_module = importlib.import_module(full_module_name)
                logging.info(f"Loading user defined step from {steps_module}")
            for name, step_cls in inspect.getmembers(steps_module):
                base_class = getattr(step_cls, "__bases__", None)
                if (
                    name.startswith("__")
                    or base_class is None
                    or len(base_class) == 0
                    or base_class[0].__name__ != "Step"
                ):
                    continue
                logging.info(f"Loading user defined step {name} from {steps_module}")
                if inspect.isclass(step_cls):
                    step_cls.scope = scope
                    is_new_class = True
                    for i, step in enumerate(self._step_classes):
                        if step.__name__ == step_cls.__name__:
                            logging.info(f"Reloading user defined step {name} class {step_cls.__name__}")
                            self._step_classes[i] = step_cls
                            is_new_class = False
                    if is_new_class:
                        logging.info(f"Loading NEW user defined step {name} class {step_cls.__name__}")
                        self._step_classes.append(step_cls)

    def run_step(
        self,
        step_name: str,
        flow_context: dict[str, DataContract],
        step_configs: dict[str, Any],
        metrics: NeatMetricsCollector | None = None,
        workflow_configs: WorkflowConfigs | None = None,
        workflow_id: str = "",
        workflow_run_id: str = "",
        step_complex_configs: dict[str, Any] | None = None,
    ) -> DataContract | tuple[FlowMessage, DataContract] | FlowMessage:
        if step_complex_configs is None:
            step_complex_configs = {}
        for step_cls in self._step_classes:
            if step_cls.__name__ == step_name:
                step_obj: Step = step_cls(config=self.config)
                step_obj.configure(step_configs, step_complex_configs)
                step_obj.set_flow_context(flow_context)
                step_obj.set_metrics(metrics)
                step_obj.set_workflow_configs(workflow_configs)
                step_obj.set_workflow_metadata(workflow_id, workflow_run_id)
                signature = inspect.signature(step_obj.run)
                parameters = signature.parameters
                is_valid = True
                input_data = []
                missing_data = []
                for parameter in parameters.values():
                    try:
                        if parameter.annotation is FlowMessage:
                            input_data.append(flow_context["FlowMessage"])
                        else:
                            if isinstance(parameter.annotation, types.UnionType):
                                for param in parameter.annotation.__args__:
                                    if (
                                        param.__name__ != "_empty"
                                        and param.__name__ in flow_context
                                        and flow_context[param.__name__] is not None
                                    ):
                                        input_data.append(flow_context[param.__name__])
                                        break  # Only one variable can be used as input
                            else:
                                input_data.append(flow_context[parameter.annotation.__name__])
                    except KeyError:
                        is_valid = False
                        logging.error(f"Missing data for step {step_name} parameter {parameter.name}")
                        missing_data.append(parameter.annotation.__name__)
                        continue
                if not is_valid:
                    raise InvalidWorkFlowError(step_name, missing_data)
                return step_obj.run(*input_data)
        raise InvalidWorkFlowError(step_name, [])

    def get_list_of_steps(self) -> list[StepMetadata]:
        steps: list[StepMetadata] = []
        for step_cls in self._step_classes:
            try:
                signature = inspect.signature(step_cls.run)
                parameters = signature.parameters
                input_data = []
                output_data = []
                for parameter in parameters.values():
                    if isinstance(parameter.annotation, types.UnionType):
                        for param in parameter.annotation.__args__:
                            if param.__name__ != "_empty":
                                input_data.append(param.__name__)
                    elif parameter.annotation.__name__ != "_empty":
                        input_data.append(parameter.annotation.__name__)
                return_annotation = signature.return_annotation
                if return_annotation:
                    if isinstance(return_annotation, types.UnionType):
                        for annotation in return_annotation.__args__:
                            output_data.append(annotation.__name__)
                    elif isinstance(return_annotation, tuple):
                        for annotation in return_annotation:
                            if isinstance(annotation, types.UnionType):
                                for annotation_l2 in annotation.__args__:
                                    output_data.append(annotation_l2.__name__)
                            else:
                                output_data.append(annotation.__name__)
                    else:
                        output_data.append(return_annotation.__name__)
                steps.append(
                    StepMetadata(
                        name=step_cls.__name__,
                        scope=step_cls.scope,
                        input=input_data,
                        description=step_cls.description,
                        category=step_cls.category,
                        output=output_data,
                        configurables=step_cls.configurables,
                        version=step_cls.version,
                        source=step_cls.source,
                        docs_url=step_cls.docs_url,
                    )
                )
            except AttributeError as e:
                logging.error(
                    f"Step {type(step_cls).__name__} does not have a run method or types can't be infered.Error: {e}"
                )

        return steps



import warnings
from typing import TYPE_CHECKING, cast

from rdflib import Namespace

from cognite.neat.rules import issues
from cognite.neat.rules.models._base import SheetList
from cognite.neat.rules.models.data_types import DataType
from cognite.neat.rules.models.domain import DomainRules
from cognite.neat.rules.models.entities import (
    ClassEntity,
    ContainerEntity,
    DMSUnknownEntity,
    ParentClassEntity,
    ReferenceEntity,
    UnknownEntity,
    ViewEntity,
    ViewPropertyEntity,
)
from cognite.neat.rules.models.information._rules import InformationRules

from ._rules import DMSMetadata, DMSProperty, DMSRules, DMSView

if TYPE_CHECKING:
    from cognite.neat.rules.models.information._rules import InformationMetadata


class _DMSRulesConverter:
    def __init__(self, dms: DMSRules):
        self.dms = dms

    def as_domain_rules(self) -> "DomainRules":
        raise NotImplementedError("DomainRules not implemented yet")

    def as_information_architect_rules(
        self,
    ) -> "InformationRules":
        from cognite.neat.rules.models.information._rules import (
            InformationClass,
            InformationProperty,
            InformationRules,
        )

        dms = self.dms.metadata

        metadata = self._convert_metadata_to_info(dms)

        classes = [
            InformationClass(
                # we do not want a version in class as we use URI for the class
                class_=ClassEntity(prefix=view.class_.prefix, suffix=view.class_.suffix),
                description=view.description,
                parent=[
                    # we do not want a version in class as we use URI for the class
                    ParentClassEntity(prefix=implemented_view.prefix, suffix=implemented_view.suffix)
                    # We only want parents in the same namespace, parent in a different namespace is a reference
                    for implemented_view in view.implements or []
                    if implemented_view.prefix == view.class_.prefix
                ],
                reference=self._get_class_reference(view),
            )
            for view in self.dms.views
        ]

        properties: list[InformationProperty] = []
        value_type: DataType | ClassEntity | str
        for property_ in self.dms.properties:
            if isinstance(property_.value_type, DataType):
                value_type = property_.value_type
            elif isinstance(property_.value_type, ViewEntity | ViewPropertyEntity):
                value_type = ClassEntity(
                    prefix=property_.value_type.prefix,
                    suffix=property_.value_type.suffix,
                )
            elif isinstance(property_.value_type, DMSUnknownEntity):
                value_type = UnknownEntity()
            else:
                raise ValueError(f"Unsupported value type: {property_.value_type.type_}")

            properties.append(
                InformationProperty(
                    # Removing version
                    class_=ClassEntity(suffix=property_.class_.suffix, prefix=property_.class_.prefix),
                    property_=property_.view_property,
                    value_type=value_type,
                    description=property_.description,
                    min_count=0 if property_.nullable or property_.nullable is None else 1,
                    max_count=float("inf") if property_.is_list or property_.nullable is None else 1,
                    reference=self._get_property_reference(property_),
                )
            )

        return InformationRules(
            metadata=metadata,
            properties=SheetList[InformationProperty](data=properties),
            classes=SheetList[InformationClass](data=classes),
            last=self.dms.last.as_information_architect_rules() if self.dms.last else None,
            reference=self.dms.reference.as_information_architect_rules() if self.dms.reference else None,
        )

    @classmethod
    def _convert_metadata_to_info(cls, metadata: DMSMetadata) -> "InformationMetadata":
        from cognite.neat.rules.models.information._rules import InformationMetadata

        prefix = metadata.space
        return InformationMetadata(
            schema_=metadata.schema_,
            data_model_type=metadata.data_model_type,
            extension=metadata.extension,
            prefix=prefix,
            namespace=Namespace(f"https://purl.orgl/neat/{prefix}/"),
            version=metadata.version,
            description=metadata.description,
            name=metadata.name or metadata.external_id,
            creator=metadata.creator,
            created=metadata.created,
            updated=metadata.updated,
        )

    @classmethod
    def _get_class_reference(cls, view: DMSView) -> ReferenceEntity | None:
        parents_other_namespace = [parent for parent in view.implements or [] if parent.prefix != view.class_.prefix]
        if len(parents_other_namespace) == 0:
            return None
        if len(parents_other_namespace) > 1:
            warnings.warn(
                issues.dms.MultipleReferenceWarning(
                    view_id=view.view.as_id(), implements=[v.as_id() for v in parents_other_namespace]
                ),
                stacklevel=2,
            )
        other_parent = parents_other_namespace[0]

        return ReferenceEntity(prefix=other_parent.prefix, suffix=other_parent.suffix)

    @classmethod
    def _get_property_reference(cls, property_: DMSProperty) -> ReferenceEntity | None:
        has_container_other_namespace = property_.container and property_.container.prefix != property_.class_.prefix
        if not has_container_other_namespace:
            return None
        container = cast(ContainerEntity, property_.container)
        return ReferenceEntity(
            prefix=container.prefix,
            suffix=container.suffix,
            property=property_.container_property,
        )



import re
from collections.abc import Callable
from typing import Annotated, Any, cast

import rdflib
from pydantic import (
    AfterValidator,
    BeforeValidator,
    Field,
    HttpUrl,
    StringConstraints,
    TypeAdapter,
    ValidationInfo,
    WrapValidator,
)
from pydantic.functional_serializers import PlainSerializer
from pydantic_core import PydanticCustomError

from cognite.neat.rules import exceptions

from ._base import (
    MORE_THAN_ONE_NONE_ALPHANUMERIC_REGEX,
    PREFIX_COMPLIANCE_REGEX,
    PROPERTY_ID_COMPLIANCE_REGEX,
    VERSION_COMPLIANCE_REGEX,
)


def _custom_error(exc_factory: Callable[[str | None, Exception], Any]) -> Any:
    def _validator(value: Any, next_: Any, ctx: ValidationInfo) -> Any:
        try:
            return next_(value, ctx)
        except Exception:
            raise exc_factory(ctx.field_name, value) from None

    return WrapValidator(_validator)


def _raise(exception: PydanticCustomError):
    raise exception


StrOrListType = Annotated[
    str | list[str],
    BeforeValidator(lambda value: value.replace(", ", ",").split(",") if isinstance(value, str) and value else value),
    PlainSerializer(
        lambda value: ",".join([entry for entry in value if entry]) if isinstance(value, list) else value,
        return_type=str,
        when_used="unless-none",
    ),
]


StrListType = Annotated[
    list[str],
    BeforeValidator(lambda value: [entry.strip() for entry in value.split(",")] if isinstance(value, str) else value),
    PlainSerializer(
        lambda value: ",".join([entry for entry in value if entry]), return_type=str, when_used="unless-none"
    ),
]

NamespaceType = Annotated[
    rdflib.Namespace,
    BeforeValidator(
        lambda value: (
            rdflib.Namespace(TypeAdapter(HttpUrl).validate_python(value))
            if value.endswith("#") or value.endswith("/")
            else rdflib.Namespace(TypeAdapter(HttpUrl).validate_python(f"{value}/"))
        )
    ),
]

PrefixType = Annotated[
    str,
    StringConstraints(pattern=PREFIX_COMPLIANCE_REGEX),
    _custom_error(
        lambda _, value: exceptions.PrefixesRegexViolation(
            cast(list[str], [value]), PREFIX_COMPLIANCE_REGEX
        ).to_pydantic_custom_error()
    ),
]

ExternalIdType = Annotated[
    str,
    Field(min_length=1, max_length=255),
]

VersionType = Annotated[
    str,
    StringConstraints(pattern=VERSION_COMPLIANCE_REGEX),
    _custom_error(
        lambda _, value: exceptions.VersionRegexViolation(
            version=cast(str, value), regex_expression=VERSION_COMPLIANCE_REGEX
        ).to_pydantic_custom_error()
    ),
]


PropertyType = Annotated[
    str,
    AfterValidator(
        lambda value: (
            _raise(exceptions.MoreThanOneNonAlphanumericCharacter("property", value).to_pydantic_custom_error())
            if re.search(MORE_THAN_ONE_NONE_ALPHANUMERIC_REGEX, value)
            else (
                value
                if re.match(PROPERTY_ID_COMPLIANCE_REGEX, value)
                else _raise(
                    exceptions.PropertyIDRegexViolation(value, PROPERTY_ID_COMPLIANCE_REGEX).to_pydantic_custom_error()
                )
            )
        )
    ),
]



import datetime
import re

import pandas as pd
from rdflib import Graph, Namespace

from cognite.neat.legacy.rules.models.rules import (
    cdf_space_compliance_regex,
    data_model_id_compliance_regex,
    prefix_compliance_regex,
    version_compliance_regex,
)
from cognite.neat.utils.utils import convert_rdflib_content, remove_none_elements_from_set


def parse_owl_metadata(graph: Graph, make_compliant: bool = False) -> pd.DataFrame:
    """Parse owl metadata from graph to pandas dataframe.

    Args:
        graph: Graph containing owl metadata
        make_compliant: Flag for generating compliant metadata, by default False

    Returns:
        Dataframe containing owl metadata

    !!! note "make_compliant"
        If `make_compliant` is set to True, in presence of errors, default values will be used instead.
        This makes the method very opinionated, but results in a compliant metadata.


    """
    # TODO: Move dataframe to dict representation

    query = """SELECT ?namespace ?prefix ?dataModelName ?cdfSpaceName ?version ?isCurrentVersion
    ?created ?updated ?title ?description ?creator ?contributor ?rights ?license
    WHERE {
        ?namespace a owl:Ontology .
        OPTIONAL {?namespace owl:versionInfo ?version }.
        OPTIONAL {?namespace dcterms:creator ?creator }.
        OPTIONAL {?namespace dcterms:title|rdfs:label|skos:prefLabel ?title }.
        OPTIONAL {?namespace dcterms:contributor ?contributor }.
        OPTIONAL {?namespace dcterms:modified ?updated }.
        OPTIONAL {?namespace dcterms:created ?created }.
        OPTIONAL {?namespace dcterms:description ?description }.

        OPTIONAL {?namespace dcterms:rights|dc:rights ?rights }.

        OPTIONAL {?namespace dcterms:license|dc:license ?license }.
        FILTER (!isBlank(?namespace))
        FILTER (!bound(?description) || LANG(?description) = "" || LANGMATCHES(LANG(?description), "en"))
        FILTER (!bound(?title) || LANG(?title) = "" || LANGMATCHES(LANG(?title), "en"))
    }
    """

    results = [{item for item in sublist} for sublist in list(zip(*graph.query(query), strict=True))]

    clean_list = convert_rdflib_content(
        {
            "namespace": Namespace(results[0].pop()),
            "prefix": results[1].pop(),
            "dataModelName": results[2].pop(),
            "cdfSpaceName": results[3].pop(),
            "version": results[4].pop(),
            "isCurrentVersion": results[5].pop(),
            "created": results[6].pop(),
            "updated": results[7].pop(),
            "title": results[8].pop(),
            "description": results[9].pop(),
            "creator": (
                ", ".join(remove_none_elements_from_set(results[10]))
                if remove_none_elements_from_set(results[10])
                else None
            ),
            "contributor": (
                ", ".join(remove_none_elements_from_set(results[11]))
                if remove_none_elements_from_set(results[11])
                else None
            ),
            "rights": results[12].pop(),
            "license": results[13].pop(),
        }
    )

    if make_compliant:
        clean_list.pop("created")
        return pd.DataFrame(list(make_metadata_compliant(clean_list).items()), columns=["Key", "Value"])

    return pd.DataFrame(list(clean_list.items()), columns=["Key", "Value"])


def make_metadata_compliant(metadata: dict) -> dict:
    """Attempts to fix errors in metadata, otherwise defaults to values that will pass validation.

    Args:
        metadata: Dictionary containing metadata

    Returns:
        Dictionary containing metadata with fixed errors
    """

    metadata = fix_namespace(metadata, default=Namespace("http://purl.org/cognite/neat#"))
    metadata = fix_prefix(metadata)
    metadata = fix_dataModelName(metadata)
    metadata = fix_cdfSpaceName(metadata)
    metadata = fix_version(metadata)
    metadata = fix_isCurrentVersion(metadata)
    metadata = fix_date(metadata, date_type="created", default=datetime.datetime.now().replace(microsecond=0))
    metadata = fix_date(metadata, date_type="updated", default=datetime.datetime.now().replace(microsecond=0))
    metadata = fix_title(metadata)
    metadata = fix_description(metadata)
    metadata = fix_author(metadata, "creator")
    metadata = fix_author(metadata, "contributor", "Cognite")
    metadata = fix_rights(metadata)
    metadata = fix_license(metadata)

    return metadata


def fix_license(metadata: dict, default: str = "Unknown license") -> dict:
    if license := metadata.get("license", None):
        if not isinstance(license, str):
            metadata["license"] = default
        elif isinstance(license, str) and len(license) == 0:
            metadata["license"] = default
    else:
        metadata["license"] = default
    return metadata


def fix_rights(metadata: dict, default: str = "Unknown rights") -> dict:
    if rights := metadata.get("rights", None):
        if not isinstance(rights, str):
            metadata["rights"] = default
        elif isinstance(rights, str) and len(rights) == 0:
            metadata["rights"] = default
    else:
        metadata["rights"] = default
    return metadata


def fix_author(metadata: dict, author_type: str = "creator", default: str = "NEAT") -> dict:
    if author := metadata.get(author_type, None):
        if not isinstance(author, str) or isinstance(author, list):
            metadata[author_type] = default
        elif isinstance(author, str) and len(author) == 0:
            metadata[author_type] = default
    else:
        metadata[author_type] = default
    return metadata


def fix_description(metadata: dict, default: str = "This model has been inferred from OWL ontology") -> dict:
    if description := metadata.get("description", None):
        if not isinstance(description, str) or len(description) == 0:
            metadata["description"] = default
        elif isinstance(description, str) and len(description) > 1024:
            metadata["description"] = metadata["description"][:1024]
    else:
        metadata["description"] = default
    return metadata


def fix_cdfSpaceName(metadata: dict, default: str = "playground") -> dict:
    if space := metadata.get("cdfSpaceName", None):
        if not isinstance(space, str) or not re.match(cdf_space_compliance_regex, space):
            metadata["cdfSpaceName"] = default
    else:
        metadata["cdfSpaceName"] = default
    return metadata


def fix_dataModelName(metadata: dict, default: str = "neat") -> dict:
    if data_model_name := metadata.get("dataModelName", None):
        if not isinstance(data_model_name, str) or not re.match(data_model_id_compliance_regex, data_model_name):
            metadata["dataModelName"] = default
    else:
        metadata["dataModelName"] = default
    return metadata


def fix_prefix(metadata: dict, default: str = "neat") -> dict:
    if prefix := metadata.get("prefix", None):
        if not isinstance(prefix, str) or not re.match(prefix_compliance_regex, prefix):
            metadata["prefix"] = default
    else:
        metadata["prefix"] = default
    return metadata


def fix_namespace(metadata: dict, default: Namespace) -> dict:
    if namespace := metadata.get("namespace", None):
        if not isinstance(namespace, Namespace):
            try:
                metadata["namespace"] = Namespace(namespace)
            except Exception:
                metadata["namespace"] = default
    else:
        metadata["namespace"] = default

    return metadata


def fix_date(
    metadata: dict,
    date_type: str,
    default: datetime.datetime,
) -> dict:
    if date := metadata.get(date_type, None):
        try:
            if isinstance(date, datetime.datetime):
                pass
            elif isinstance(date, datetime.date):
                metadata[date_type] = datetime.datetime.combine(metadata[date_type], datetime.datetime.min.time())
            elif isinstance(date, str):
                metadata[date_type] = datetime.datetime.strptime(metadata[date_type], "%Y-%m-%dT%H:%M:%SZ")
            else:
                metadata[date_type] = default
        except Exception:
            metadata[date_type] = default
    else:
        metadata[date_type] = default

    return metadata


def fix_version(metadata: dict, default: str = "1.0.0") -> dict:
    if version := metadata.get("version", None):
        if not re.match(version_compliance_regex, version):
            metadata["version"] = default
    else:
        metadata["version"] = default

    return metadata


def fix_isCurrentVersion(metadata: dict, default: bool = True) -> dict:
    if isCurrentVersion := metadata.get("isCurrentVersion", None):
        if not isinstance(isCurrentVersion, bool):
            metadata["isCurrentVersion"] = default
    else:
        metadata["isCurrentVersion"] = default

    return metadata


def fix_title(metadata: dict, default: str = "OWL Inferred Data Model") -> dict:
    if title := metadata.get("title", None):
        if not isinstance(title, str):
            metadata["title"] = default
        elif isinstance(title, str) and len(title) == 0:
            metadata["title"] = default
        elif isinstance(title, str) and len(title) > 255:
            metadata["title"] = metadata["title"][:255]
        else:
            pass
    else:
        metadata["title"] = default

    return metadata



import logging

from rdflib import Graph, Namespace

from cognite.neat.constants import DEFAULT_NAMESPACE, PREFIXES

from ._base import NeatGraphStoreBase


class MemoryStore(NeatGraphStoreBase):
    """MemoryStore is a class that stores the graph in memory using rdflib and provides
    methods to read/write data it contains.


    Args:
        graph : Instance of rdflib.Graph class for graph storage
        base_prefix : Used as a base prefix for graph namespace, allowing querying graph data using a shortform of a URI
        namespace : Namespace (aka URI) used to resolve any relative URI in the graph
        prefixes : Dictionary of additional prefixes used and bounded to the graph
    """

    rdf_store_type: str = "memory"

    def __init__(
        self,
        graph: Graph | None = None,
        base_prefix: str = "",  # usually empty
        namespace: Namespace = DEFAULT_NAMESPACE,
        prefixes: dict = PREFIXES,
    ):
        # Init repeated to get nice docstring
        super().__init__(graph, base_prefix, namespace, prefixes)

    def _set_graph(self):
        logging.info("Initializing graph in memory")
        self.graph = Graph()

    def drop(self):
        """Drops the graph."""
        # In the case of in-memory graph, we just reinitialize the graph
        # otherwise we would lose the prefixes and bindings, which fails
        # workflow
        self.reinitialize_graph()



from ._base import NeatGraphStoreBase
from ._graphdb_store import GraphDBStore
from ._memory_store import MemoryStore
from ._oxigraph_store import OxiGraphStore

STORE_BY_TYPE: dict[str, type[NeatGraphStoreBase]] = {}
for store in NeatGraphStoreBase.__subclasses__():
    STORE_BY_TYPE[store.rdf_store_type] = store  # type: ignore[type-abstract]

del store  # Cleanup namespace
AVAILABLE_STORES = set(STORE_BY_TYPE.keys())

__all__ = ["NeatGraphStoreBase", "MemoryStore", "OxiGraphStore", "GraphDBStore", "STORE_BY_TYPE", "AVAILABLE_STORES"]



import sys
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Generic, TypeVar

from cognite.neat.legacy.rules.models.raw_rules import RawRules
from cognite.neat.legacy.rules.models.rules import Rules

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

T_Export = TypeVar("T_Export")


class BaseExporter(ABC, Generic[T_Export]):
    def __init__(self, rules: Rules, report: str | None = None):
        self.rules = rules
        self.report = report

    @classmethod
    def from_rules(cls, rules: RawRules | Rules) -> Self:
        if isinstance(rules, Rules):
            return cls(rules)
        elif isinstance(rules, RawRules):
            validated_rules = rules.to_rules(skip_validation=True)
            report = rules.validate_rules()
            return cls(validated_rules, report)

        raise ValueError(f"Expected RawRules or Rules, got {type(rules)}")

    def export_to_file(self, filepath: Path, report_path: Path | None = None) -> None:
        self._export_to_file(filepath)
        if self.report:
            report_path = report_path or filepath.parent / f"{filepath.stem}_validation_report.txt"
            report_path.write_text(self.report)

    @abstractmethod
    def _export_to_file(self, filepath: Path) -> None:
        raise NotImplementedError

    @abstractmethod
    def export(self) -> T_Export:
        raise NotImplementedError



import logging
import time
from datetime import datetime
from pathlib import Path
from typing import Any, ClassVar, cast

from cognite.client import CogniteClient
from cognite.client.data_classes import Asset, AssetFilter
from prometheus_client import Gauge

from cognite.neat.legacy.graph import loaders as graph_loader
from cognite.neat.legacy.graph.loaders import upload_labels
from cognite.neat.legacy.graph.loaders.core.rdf_to_assets import (
    NeatMetadataKeys,
    categorize_assets,
    rdf2assets,
    remove_non_existing_labels,
    unique_asset_labels,
    upload_assets,
)
from cognite.neat.legacy.graph.loaders.core.rdf_to_relationships import (
    categorize_relationships,
    rdf2relationships,
    upload_relationships,
)
from cognite.neat.legacy.graph.loaders.rdf_to_dms import upload_edges, upload_nodes
from cognite.neat.legacy.graph.loaders.validator import validate_asset_hierarchy
from cognite.neat.legacy.rules.models.rdfpath import TransformationRuleType
from cognite.neat.utils.utils import generate_exception_report
from cognite.neat.workflows._exceptions import StepFlowContextNotInitialized, StepNotInitialized
from cognite.neat.workflows.model import FlowMessage, StepExecutionStatus
from cognite.neat.workflows.steps.data_contracts import (
    CategorizedAssets,
    CategorizedRelationships,
    Edges,
    Nodes,
    RulesData,
    SolutionGraph,
    SourceGraph,
)
from cognite.neat.workflows.steps.step_model import Configurable, Step

__all__ = [
    "GenerateAssetsFromGraph",
    "GenerateRelationshipsFromGraph",
    "GenerateNodesAndEdgesFromGraph",
    "LoadLabelsToCDF",
    "LoadAssetsToCDF",
    "LoadRelationshipsToCDF",
    "LoadNodesToCDF",
    "LoadEdgesToCDF",
    "LoadGraphToRdfFile",
]

CATEGORY = __name__.split(".")[-1].replace("_", " ").title() + " [LEGACY]"


class LoadLabelsToCDF(Step):
    """
    This step creates and loads default NEAT labels in CDF
    """

    description = "This step creates default NEAT labels in CDF"
    category = CATEGORY
    version = "legacy"
    configurables: ClassVar[list[Configurable]] = [
        Configurable(name="data_set_id", value="", label=("CDF dataset id to which the labels will be added."))
    ]

    def run(self, rules: RulesData, cdf_client: CogniteClient) -> None:  # type: ignore[override, syntax]
        upload_labels(
            cdf_client,
            rules.rules,
            data_set_id=int(self.configs["data_set_id"]),
            extra_labels=["non-historic", "historic"],
        )


class GenerateNodesAndEdgesFromGraph(Step):
    """
    The step generates nodes and edges from the graph
    """

    description = "The step generates nodes and edges from the graph"
    category = CATEGORY

    configurables: ClassVar[list[Configurable]] = [
        Configurable(
            name="graph_name",
            value="source",
            options=["source", "solution"],
            label=("The name of the graph to be used for matching." " Supported options : source, solution"),
        ),
        Configurable(
            name="add_class_prefix",
            value="False",
            options=["True", "False"],
            label=("Whether to add class name as a prefix to external ids of instances or not"),
        ),
        Configurable(
            name="data_validation_error_handling_strategy",
            value="skip_and_report",
            options=["skip_and_report", "fail_and_report"],
            label=(
                "The strategy for handling data validation errors. Supported options: \
                   skip_and_report - failed instance (node or edge) will be skipped and reported , \
                   fail_and_report - failed instance  (node or edge) will fail the workflow and report the error"
            ),
        ),
        Configurable(
            name="apply_basic_transformation",
            value="True",
            options=["True", "False"],
            label=("Whether to apply basic transformations rules (rdfpath) or not. Default is True."),
        ),
    ]

    def run(  # type: ignore[override, syntax]
        self, rules: RulesData, graph: SourceGraph | SolutionGraph
    ) -> (FlowMessage, Nodes, Edges):  # type: ignore[syntax]
        if self.configs is None or self.data_store_path is None:
            raise StepNotInitialized(type(self).__name__)
        if self.flow_context is None:
            raise StepFlowContextNotInitialized(type(self).__name__)

        graph_name = self.configs["graph_name"] or "source"
        data_validation_error_handling_strategy = self.configs.get(
            "data_validation_error_handling_strategy", "skip_and_report"
        )
        if graph_name == "solution":
            # Todo Anders: Why is the graph fetched from context when it is passed as an argument?
            graph = cast(SourceGraph | SolutionGraph, self.flow_context["SolutionGraph"])
        else:
            graph = cast(SourceGraph | SolutionGraph, self.flow_context["SourceGraph"])

        add_class_prefix = True if self.configs["add_class_prefix"] == "True" else False
        apply_basic_transformation = True if self.configs.get("apply_basic_transformation", "True") == "True" else False

        if apply_basic_transformation:
            final_rules = rules.rules
        else:
            logging.debug("Basic transformation rules are not applied to the graph")
            final_rules = rules.rules.model_copy(deep=True)
            prefix = final_rules.metadata.prefix
            for rule in final_rules.properties.values():
                rule.rule_type = TransformationRuleType.rdfpath
                rule.rule = f"{prefix}:{rule.class_id}({prefix}:{rule.property_id})"

        loader = graph_loader.DMSLoader(final_rules, graph.graph, add_class_prefix=add_class_prefix)
        nodes, edges, exceptions = loader.as_nodes_and_edges(stop_on_exception=False)

        msg = f"Total count of: <ul><li>{ len(nodes) } nodes</li><li>{ len(edges) } edges</li></ul>"

        if exceptions:
            file_name = f'nodes-and-edges-exceptions_{datetime.now().strftime("%Y%d%m%H%M")}.txt'
            exceptions_report_dir = self.data_store_path / "reports"
            exceptions_report_dir.mkdir(parents=True, exist_ok=True)
            exceptions_report_path = exceptions_report_dir / file_name

            exceptions_report_path.write_text(generate_exception_report(exceptions, "Errors"))
            msg += (
                f"<p>There is total of { len(exceptions) } exceptions</p>"
                f'<a href="/data/reports/{file_name}?{time.time()}" '
                f'target="_blank">Full error report </a>'
            )
            if data_validation_error_handling_strategy == "fail_and_report":
                return FlowMessage(error_text=msg, step_execution_status=StepExecutionStatus.ABORT_AND_FAIL)

        return FlowMessage(output_text=msg), Nodes(nodes=nodes), Edges(edges=edges)


class LoadGraphToRdfFile(Step):
    """
    The step generates loads graph to RDF file
    """

    description = "The step generates nodes and edges from the graph"
    category = CATEGORY
    version = "legacy"
    configurables: ClassVar[list[Configurable]] = [
        Configurable(
            name="graph_name",
            value="source",
            options=["source", "solution"],
            label=("The name of the graph to be used for loading RDF File." " Supported options : source, solution"),
        ),
        Configurable(
            name="rdf_file_path",
            value="staging/graph_export.ttl",
            label=("Relative path for the RDF file storage, " "must end with .ttl !"),
        ),
    ]

    def run(  # type: ignore[override, syntax]
        self, graph: SourceGraph | SolutionGraph
    ) -> FlowMessage:  # type: ignore[syntax]
        if self.configs is None or self.data_store_path is None:
            raise StepNotInitialized(type(self).__name__)

        storage_path = self.data_store_path / Path(self.configs["rdf_file_path"])
        relative_graph_file_path = str(storage_path).split("/data/")[1]

        graph_name = self.configs["graph_name"] or "source"

        if graph_name == "solution":
            # Todo Anders: Why is the graph fetched from context when it is passed as an argument?
            graph = cast(SourceGraph | SolutionGraph, self.flow_context["SolutionGraph"])
        else:
            graph = cast(SourceGraph | SolutionGraph, self.flow_context["SourceGraph"])

        graph.graph.serialize(str(storage_path), format="turtle")

        output_text = (
            "<p></p>"
            "Graph loaded to RDF file can be downloaded here : "
            f'<a href="/data/{relative_graph_file_path}?{time.time()}" '
            f'target="_blank">{storage_path.stem}.ttl</a>'
        )

        return FlowMessage(output_text=output_text)


class LoadNodesToCDF(Step):
    """
    This step uploads nodes to CDF
    """

    description = "This step uploads nodes to CDF"
    category = CATEGORY
    version = "legacy"

    def run(self, cdf_client: CogniteClient, nodes: Nodes) -> FlowMessage:  # type: ignore[override, syntax]
        if nodes.nodes:
            upload_nodes(cdf_client, nodes.nodes, max_retries=2, retry_delay=4)
            return FlowMessage(output_text="CDF nodes uploaded successfully")
        else:
            return FlowMessage(output_text="No nodes to upload!")


class LoadEdgesToCDF(Step):
    """
    This step uploads edges to CDF
    """

    description = "This step uploads edges to CDF"
    category = CATEGORY
    version = "legacy"

    def run(self, cdf_client: CogniteClient, edges: Edges) -> FlowMessage:  # type: ignore[override, syntax]
        if edges.edges:
            upload_edges(cdf_client, edges.edges, max_retries=2, retry_delay=4)
            return FlowMessage(output_text="CDF edges uploaded successfully")
        else:
            return FlowMessage(output_text="No edges to upload!")


class GenerateAssetsFromGraph(Step):
    """
    The step generates assets from the graph ,categorizes them and stores them in CategorizedAssets object
    """

    description = (
        "The step generates assets from the graph ,categorizes them and stores them in CategorizedAssets object"
    )
    category = CATEGORY
    version = "legacy"
    configurables: ClassVar[list[Configurable]] = [
        Configurable(name="data_set_id", value="", label=("CDF dataset id to which the labels will be added.")),
        Configurable(
            name="asset_external_id_prefix",
            value="",
            label=("Prefix to be added to all asset external ids, default None."),
        ),
        Configurable(
            name="assets_cleanup_type",
            value="nothing",
            options=["nothing", "orphans", "circular", "full"],
            label=(
                "Configures asset cleanup process. Supported options: nothing - no cleanup, \
                    orphans - all orphan assets will be removed, circular - all circular assets will be removed , \
                    full - full cleanup , both orphans and circular assets will be removed. "
            ),
        ),
    ]

    def run(  # type: ignore[override]
        self, rules: RulesData, cdf_client: CogniteClient, solution_graph: SolutionGraph
    ) -> (FlowMessage, CategorizedAssets):  # type: ignore[override, syntax]
        if self.configs is None:
            raise StepNotInitialized(type(self).__name__)
        asset_cleanup_type = self.configs.get("assets_cleanup_type", "nothing")
        data_set_id = int(self.configs["data_set_id"])
        asset_external_id_prefix = self.configs.get("asset_external_id_prefix", None)

        meta_keys = NeatMetadataKeys.load(self.configs)

        if self.metrics is None:
            raise ValueError(self._not_configured_message)
        prom_cdf_resource_stats = cast(
            Gauge,
            self.metrics.register_metric(
                "cdf_resources_stats",
                "CDF resource stats before and after running the workflow",
                m_type="gauge",
                metric_labels=["resource_type", "state"],
            ),
        )
        prom_data_issues_stats = cast(
            Gauge,
            self.metrics.register_metric(
                "data_issues_stats", "Data validation issues", m_type="gauge", metric_labels=["resource_type"]
            ),
        )

        rdf_asset_dicts = rdf2assets(
            solution_graph.graph,
            rules.rules,
            data_set_id=data_set_id,
            asset_external_id_prefix=asset_external_id_prefix,
            stop_on_exception=True,
            meta_keys=meta_keys,
        )

        # UPDATE: 2023-04-05 - correct aggregation of assets in CDF for specific dataset
        total_assets_before = cdf_client.assets.aggregate(filter=AssetFilter(data_set_ids=[{"id": data_set_id}]))[
            0
        ].count

        # Label Validation
        labels_before = unique_asset_labels(rdf_asset_dicts.values())
        logging.info(f"Assets have {len(labels_before)} unique labels: {', '.join(sorted(labels_before))}")

        rdf_asset_dicts = cast(dict[str, dict[str, Any]], remove_non_existing_labels(cdf_client, rdf_asset_dicts))

        labels_after = unique_asset_labels(rdf_asset_dicts.values())
        removed_labels = labels_before - labels_after
        logging.info(
            f"Removed {len(removed_labels)} labels as these do not exists in CDF. "
            f"Removed labels: {', '.join(sorted(removed_labels))}"
        )
        ######################

        prom_cdf_resource_stats.labels(resource_type="asset", state="count_before_neat_update").set(total_assets_before)
        logging.info(f"Total count of assets in CDF before upload: { total_assets_before }")

        orphanage_asset_external_id = (
            f"{asset_external_id_prefix}orphanage-{data_set_id}"
            if asset_external_id_prefix
            else f"orphanage-{data_set_id}"
        )
        orphan_assets, circular_assets, parent_children_map = validate_asset_hierarchy(rdf_asset_dicts)

        # There could be assets already under a created orphan assets. Include those in oprhan assets list
        if orphanage_asset_external_id in parent_children_map:
            orphan_assets.extend(parent_children_map[orphanage_asset_external_id])

        orphan_assets_count = len(orphan_assets)
        circular_assets_count = len(circular_assets)
        prom_data_issues_stats.labels(resource_type="circular_assets").set(len(circular_assets))
        prom_data_issues_stats.labels(resource_type="orphan_assets").set(len(orphan_assets))

        if orphan_assets:
            logging.error(f"Found orphaned assets: {', '.join(orphan_assets)}")

            if asset_cleanup_type in ["orphans", "full"]:
                logging.info("Removing orphaned assets and its children")

                def delete_asset_and_children_recursive(asset_id, rdf_asset_dicts, parent_children_map):
                    if asset_id in rdf_asset_dicts:
                        del rdf_asset_dicts[asset_id]

                    if asset_id in parent_children_map:
                        for child_id in parent_children_map[asset_id]:
                            delete_asset_and_children_recursive(child_id, rdf_asset_dicts, parent_children_map)

                def delete_orphan_assets_recursive(orphan_assets, rdf_asset_dicts, parent_children_map):
                    for orphan_asset in orphan_assets:
                        delete_asset_and_children_recursive(orphan_asset, rdf_asset_dicts, parent_children_map)

                # Make sure children, grand-children, great-grandchildren .... are deleted
                delete_orphan_assets_recursive(orphan_assets, rdf_asset_dicts, parent_children_map)

                # delete orphange asset
                if orphanage_asset_external_id in rdf_asset_dicts:
                    del rdf_asset_dicts[orphanage_asset_external_id]

            else:
                # Kill the process if you dont have orphanage asset in your asset hierarchy
                # and inform the user that it is missing !
                if orphanage_asset_external_id not in rdf_asset_dicts:
                    msg = f"You dont have Orphanage asset {orphanage_asset_external_id} in asset hierarchy!"
                    logging.error(msg)
                    return FlowMessage(
                        error_text=msg, step_execution_status=StepExecutionStatus.ABORT_AND_FAIL
                    ), CategorizedAssets(assets={})

                logging.error("Orphaned assets will be assigned to 'Orphanage' root asset")

                for external_id in orphan_assets:
                    rdf_asset_dicts[external_id]["parent_external_id"] = orphanage_asset_external_id
        else:
            logging.info("No orphaned assets found, your assets look healthy !")

        if circular_assets:
            logging.error(f"Found circular dependencies: {circular_assets}")
            if asset_cleanup_type in ["circular", "full"]:
                logging.info("Removing circular assets")
                for circular_path in circular_assets:
                    circular_external_id = circular_path[-1]
                    del rdf_asset_dicts[circular_external_id]
        else:
            logging.info("No circular dependency among assets found, your assets hierarchy look healthy !")

        if orphan_assets or circular_assets:
            orphan_assets, circular_assets, _ = validate_asset_hierarchy(rdf_asset_dicts)
            if circular_assets:
                msg = f"Found circular dependencies: {circular_assets!s}"
                logging.error(msg)
                return FlowMessage(
                    error_text=msg, step_execution_status=StepExecutionStatus.ABORT_AND_FAIL
                ), CategorizedAssets(assets={})
            elif orphan_assets:
                msg = f"Not able to fix orphans: {', '.join(orphan_assets)}"
                logging.error(msg)
                return FlowMessage(
                    error_text=msg, step_execution_status=StepExecutionStatus.ABORT_AND_FAIL
                ), CategorizedAssets(assets={})
            else:
                logging.info("No circular dependency among assets found, your assets hierarchy look healthy !")

        categorized_assets, report = categorize_assets(
            cdf_client, rdf_asset_dicts, data_set_id=data_set_id, return_report=True
        )

        count_create_assets = len(categorized_assets["create"])
        count_update_assets = len(categorized_assets["update"])
        count_decommission_assets = len(categorized_assets["decommission"])
        count_resurrect_assets = len(categorized_assets["resurrect"])

        prom_cdf_resource_stats.labels(resource_type="asset", state="create").set(count_create_assets)
        prom_cdf_resource_stats.labels(resource_type="asset", state="update").set(count_update_assets)
        prom_cdf_resource_stats.labels(resource_type="asset", state="decommission").set(count_decommission_assets)
        prom_cdf_resource_stats.labels(resource_type="asset", state="resurrect").set(count_resurrect_assets)

        logging.info(f"Total count of assets to be created: { count_create_assets }")
        logging.info(f"Total count of assets to be updated: { count_update_assets }")
        logging.info(f"Total count of assets to be decommission: { count_decommission_assets }")
        logging.info(f"Total count of assets to be resurrect: { count_resurrect_assets }")

        msg = f"Total count of assets { len(rdf_asset_dicts) } of which:"
        msg += f"<p> { count_create_assets } to be created </p>"
        msg += f"<p> { count_update_assets } to be updated </p>"
        msg += f"<p> { count_decommission_assets } to be decommissioned </p>"
        msg += f"<p> { count_resurrect_assets } to be resurrected </p>"
        msg += f"<p> Found { orphan_assets_count } orphan assets and"
        msg += f" { circular_assets_count } circular assets </p>"
        if asset_cleanup_type != "nothing":
            msg += " <p> All circular and orphan assets were removed successfully </p>"
        number_of_updates = len(report["decommission"])
        logging.info(f"Total number of updates: {number_of_updates}")

        return FlowMessage(output_text=msg), CategorizedAssets(assets=categorized_assets)


class LoadAssetsToCDF(Step):
    """
    This step uploads categorized assets to CDF
    """

    description = "This step uploads categorized assets to CDF"
    category = CATEGORY
    version = "legacy"

    def run(  # type: ignore[override]
        self, cdf_client: CogniteClient, categorized_assets: CategorizedAssets, flow_msg: FlowMessage
    ) -> FlowMessage:
        if flow_msg and flow_msg.payload and "action" in flow_msg.payload:
            if flow_msg.payload["action"] != "approve":
                raise Exception("Update not approved")
        if self.metrics is None:
            raise ValueError(self._not_configured_message)

        prom_cdf_resource_stats = cast(
            Gauge,
            self.metrics.register_metric(
                "cdf_resources_stats",
                "CDF resource stats before and after running the workflow",
                m_type="gauge",
                metric_labels=["resource_type", "state"],
            ),
        )
        upload_assets(cdf_client, categorized_assets.assets, max_retries=2, retry_delay=4)
        count_create_assets = len(categorized_assets.assets["create"])

        # gets first asset available irrespective of its category
        asset_example = next((assets[0] for assets in categorized_assets.assets.values() if assets), None)

        if asset_example:
            data_set_id = cast(Asset, asset_example).data_set_id
            for _ in range(1000):
                total_assets_after = cdf_client.assets.aggregate(
                    filter=AssetFilter(data_set_ids=[{"id": data_set_id}])
                )[0].count
                if total_assets_after >= count_create_assets:
                    break
                logging.info(f"Waiting for assets to be created, current count {total_assets_after}")
                time.sleep(2)

            # UPDATE: 2023-04-05 - correct aggregation of assets in CDF for specific dataset
            total_assets_after = cdf_client.assets.aggregate(filter=AssetFilter(data_set_ids=[{"id": data_set_id}]))[
                0
            ].count

            prom_cdf_resource_stats.labels(resource_type="asset", state="count_after_neat_update").set(
                total_assets_after
            )
            logging.info(f"Total count of assets in CDF after update: { total_assets_after }")
            del categorized_assets.assets  # free up memory after upload .
            return FlowMessage(output_text=f"Total count of assets in CDF after update: { total_assets_after }")
        else:
            return FlowMessage(output_text="No assets to upload!")


class GenerateRelationshipsFromGraph(Step):
    """
    This step generates relationships from the graph and saves them to CategorizedRelationships object
    """

    description = "This step generates relationships from the graph and saves them to CategorizedRelationships object"
    category = CATEGORY
    version = "legacy"
    configurables: ClassVar[list[Configurable]] = [
        Configurable(name="data_set_id", value="", label=("CDF dataset id to which the labels will be added.")),
        Configurable(
            name="relationship_external_id_prefix",
            value="",
            label=("Prefix to be added to all asset external ids, default None."),
        ),
    ]

    def run(  # type: ignore[override]
        self, rules: RulesData, cdf_client: CogniteClient, solution_graph: SolutionGraph
    ) -> (FlowMessage, CategorizedRelationships):  # type: ignore[arg-type, syntax]
        # create, categorize and upload relationships
        data_set_id = int(self.configs["data_set_id"])
        relationship_external_id_prefix = self.configs.get("relationship_external_id_prefix", None)

        rdf_relationships = rdf2relationships(
            solution_graph.graph,
            rules.rules,
            data_set_id=data_set_id,
            relationship_external_id_prefix=relationship_external_id_prefix,
        )

        categorized_relationships = categorize_relationships(cdf_client, rdf_relationships, data_set_id)
        count_defined_relationships = len(rdf_relationships)
        count_create_relationships = len(categorized_relationships["create"])
        count_decommission_relationships = len(categorized_relationships["decommission"])
        count_resurrect_relationships = len(categorized_relationships["resurrect"])

        if self.metrics is None:
            raise ValueError(self._not_configured_message)

        prom_cdf_resource_stats = cast(
            Gauge,
            self.metrics.register_metric(
                "cdf_resources_stats",
                "CDF resource stats before and after running the workflow",
                m_type="gauge",
                metric_labels=["resource_type", "state"],
            ),
        )

        prom_cdf_resource_stats.labels(resource_type="relationships", state="defined").set(count_defined_relationships)
        prom_cdf_resource_stats.labels(resource_type="relationships", state="create").set(count_create_relationships)
        prom_cdf_resource_stats.labels(resource_type="relationships", state="decommission").set(
            count_decommission_relationships
        )
        prom_cdf_resource_stats.labels(resource_type="relationships", state="resurrect").set(
            count_resurrect_relationships
        )

        msg = (
            f"Total count of relationships { count_defined_relationships } of which:"
            f" { count_create_relationships } to be created"
        )
        msg += f", { count_decommission_relationships } to be decommissioned"
        msg += f", { count_resurrect_relationships } to be resurrected"

        return FlowMessage(output_text=msg), CategorizedRelationships(relationships=categorized_relationships)


class LoadRelationshipsToCDF(Step):
    """
    This step uploads relationships to CDF
    """

    description = "This step uploads relationships to CDF"
    category = CATEGORY
    version = "legacy"

    def run(  # type: ignore[override, syntax]
        self, client: CogniteClient, categorized_relationships: CategorizedRelationships
    ) -> FlowMessage:
        upload_relationships(client, categorized_relationships.relationships, max_retries=2, retry_delay=4)
        return FlowMessage(output_text="CDF relationships uploaded successfully")



import logging
import os
from pathlib import Path

from fastapi import APIRouter

from cognite.neat.app.api.configuration import NEAT_APP
from cognite.neat.config import Config

router = APIRouter()


@router.get("/api/configs/global")
def get_configs():
    return NEAT_APP.config.dict()


@router.post("/api/configs/global")
def set_configs(request: Config):
    logging.info(f"Updating global config: {request}")
    config = request
    config.to_yaml(Path(os.environ.get("NEAT_CONFIG_PATH", "config.yaml")))
    NEAT_APP.stop()
    NEAT_APP.start(config=config)
    return config
